###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

'''
--------------------------------------------------------------

Before running this Airflow module...

  Install StarThinker in cloud composer ( recommended ):

    From Release: pip install starthinker
    From Open Source: pip install git+https://github.com/google/starthinker

  Or push local code to the cloud composer plugins directory ( if pushing local code changes ):

    source install/deploy.sh
    4) Composer Menu
    l) Install All

--------------------------------------------------------------

  If any recipe task has "auth" set to "user" add user credentials:

    1. Ensure an RECIPE['setup']['auth']['user'] = [User Credentials JSON]

  OR

    1. Visit Airflow UI > Admin > Connections.
    2. Add an Entry called "starthinker_user", fill in the following fields. Last step paste JSON from authentication.
      - Conn Type: Google Cloud Platform
      - Project: Get from https://github.com/google/starthinker/blob/master/tutorials/cloud_project.md
      - Keyfile JSON: Get from: https://github.com/google/starthinker/blob/master/tutorials/deploy_commandline.md#optional-setup-user-credentials

--------------------------------------------------------------

  If any recipe task has "auth" set to "service" add service credentials:

    1. Ensure an RECIPE['setup']['auth']['service'] = [Service Credentials JSON]

  OR

    1. Visit Airflow UI > Admin > Connections.
    2. Add an Entry called "starthinker_service", fill in the following fields. Last step paste JSON from authentication.
      - Conn Type: Google Cloud Platform
      - Project: Get from https://github.com/google/starthinker/blob/master/tutorials/cloud_project.md
      - Keyfile JSON: Get from: https://github.com/google/starthinker/blob/master/tutorials/cloud_service.md

--------------------------------------------------------------

Deal Finder

Compares open vs. deal CPM, CPC, and CPA so that clients can decide which sites, inventory, and deals work best.

  - Wait for <b>BigQuery->StarThinker Data->UNDEFINED->Deal_Finder_Dashboard</b> to be created.
  - Join the <a hre='https://groups.google.com/d/forum/starthinker-assets' target='_blank'>StarThinker Assets Group</a> to access the following assets
  - Copy <a href='https://datastudio.google.com/open/1QrWNTurvQT6nx20vnzdDveSzSmRjqHxQ' target='_blank'>Deal Finder Sample Data</a>.
  - Click Edit Connection, and change to <b>BigQuery->StarThinker Data->(field:recipe_slug}->Deal_Finder_Dashboard</b>.
  - Copy <a href='https://datastudio.google.com/open/1fjRI5AIKTYTA4fWs-pYkJbIMgCumlMyO' target='_blank'>Deal Finder Sample Report</a>.
  - When prompted choose the new data source you just created.
  - Or give these intructions to the client.

--------------------------------------------------------------

This StarThinker DAG can be extended with any additional tasks from the following sources:
  - https://google.github.io/starthinker/
  - https://github.com/google/starthinker/tree/master/dags

'''

from starthinker.airflow.factory import DAG_Factory

INPUTS = {
  'recipe_slug': '',  # Place where tables will be written in BigQuery.
  'recipe_timezone': 'America/Los_Angeles',  # Timezone for report dates.
  'recipe_name': '',  # Name of report in DV360, should be unique.
  'auth_write': 'service',  # Credentials used for writing data.
  'auth_read': 'user',  # Credentials used for reading data.
  'partners': [],  # DV360 partner id.
  'advertisers': [],  # Comma delimited list of DV360 advertiser ids.
}

RECIPE = {
  'setup': {
    'day': [
      'Mon',
      'Tue',
      'Wed',
      'Thu',
      'Fri',
      'Sat',
      'Sun'
    ],
    'hour': [
      3,
      4
    ]
  },
  'tasks': [
    {
      'dataset': {
        'description': 'Create a dataset for bigquery tables.',
        'hour': [
          4
        ],
        'auth': {
          'field': {
            'name': 'auth_write',
            'kind': 'authentication',
            'order': 1,
            'default': 'service',
            'description': 'Credentials used for writing data.'
          }
        },
        'dataset': {
          'field': {
            'name': 'recipe_slug',
            'kind': 'string',
            'description': 'Place where tables will be created in BigQuery.'
          }
        }
      }
    },
    {
      'dbm': {
        'description': 'Create a DV360 report.',
        'hour': [
          3
        ],
        'auth': {
          'field': {
            'name': 'auth_read',
            'kind': 'authentication',
            'order': 1,
            'default': 'user',
            'description': 'Credentials used for reading data.'
          }
        },
        'report': {
          'filters': {
            'FILTER_PARTNER': {
              'values': {
                'field': {
                  'name': 'partners',
                  'kind': 'integer_list',
                  'order': 5,
                  'default': [
                  ],
                  'description': 'DV360 partner id.'
                }
              }
            },
            'FILTER_ADVERTISER': {
              'values': {
                'field': {
                  'name': 'advertisers',
                  'kind': 'integer_list',
                  'order': 6,
                  'default': [
                  ],
                  'description': 'Comma delimited list of DV360 advertiser ids.'
                }
              }
            }
          },
          'body': {
            'timezoneCode': {
              'field': {
                'name': 'recipe_timezone',
                'kind': 'timezone',
                'description': 'Timezone for report dates.',
                'default': 'America/Los_Angeles'
              }
            },
            'metadata': {
              'title': {
                'field': {
                  'name': 'recipe_name',
                  'kind': 'string',
                  'prefix': 'Deal Finder For ',
                  'description': 'Name of report in DV360, should be unique.'
                }
              },
              'dataRange': 'LAST_30_DAYS',
              'format': 'CSV'
            },
            'params': {
              'type': 'TYPE_CROSS_PARTNER',
              'groupBys': [
                'FILTER_PARTNER_NAME',
                'FILTER_PARTNER',
                'FILTER_ADVERTISER_NAME',
                'FILTER_ADVERTISER',
                'FILTER_APP_URL',
                'FILTER_SITE_ID',
                'FILTER_INVENTORY_SOURCE_NAME',
                'FILTER_INVENTORY_SOURCE',
                'FILTER_INVENTORY_SOURCE_TYPE',
                'FILTER_ADVERTISER_CURRENCY',
                'FILTER_CREATIVE_WIDTH',
                'FILTER_CREATIVE_HEIGHT',
                'FILTER_CREATIVE_TYPE'
              ],
              'metrics': [
                'METRIC_IMPRESSIONS',
                'METRIC_CLICKS',
                'METRIC_TOTAL_CONVERSIONS',
                'METRIC_TOTAL_MEDIA_COST_ADVERTISER',
                'METRIC_REVENUE_ADVERTISER',
                'METRIC_ACTIVE_VIEW_MEASURABLE_IMPRESSIONS',
                'METRIC_ACTIVE_VIEW_VIEWABLE_IMPRESSIONS'
              ]
            }
          }
        }
      }
    },
    {
      'dbm': {
        'description': 'Copy a DV360 report to BigQuery.',
        'hour': [
          4
        ],
        'auth': {
          'field': {
            'name': 'auth_read',
            'kind': 'authentication',
            'order': 1,
            'default': 'user',
            'description': 'Credentials used for reading data.'
          }
        },
        'report': {
          'name': {
            'field': {
              'name': 'recipe_name',
              'kind': 'string',
              'prefix': 'Deal Finder For ',
              'description': 'Name of report in DV360, should be unique.'
            }
          },
          'timeout': 10
        },
        'out': {
          'bigquery': {
            'dataset': {
              'field': {
                'name': 'recipe_slug',
                'kind': 'string',
                'description': 'Place where tables will be written in BigQuery.'
              }
            },
            'table': 'Deal_Finder_DV360_Report',
            'header': True,
            'schema': [
              {
                'name': 'Partner',
                'type': 'STRING'
              },
              {
                'name': 'Partner_ID',
                'type': 'INTEGER'
              },
              {
                'name': 'Advertiser',
                'type': 'STRING'
              },
              {
                'name': 'Advertiser_ID',
                'type': 'INTEGER'
              },
              {
                'name': 'Site',
                'type': 'STRING'
              },
              {
                'name': 'Site_ID',
                'type': 'INTEGER'
              },
              {
                'name': 'Inventory',
                'type': 'STRING',
                'mode': 'NULLABLE'
              },
              {
                'name': 'Inventory_ID',
                'type': 'INTEGER',
                'mode': 'NULLABLE'
              },
              {
                'name': 'Inventory_Type',
                'type': 'STRING'
              },
              {
                'name': 'Advertiser_Currency',
                'type': 'STRING'
              },
              {
                'name': 'Creative_Width',
                'type': 'STRING',
                'mode': 'NULLABLE'
              },
              {
                'name': 'Creative_Height',
                'type': 'STRING',
                'mode': 'NULLABLE'
              },
              {
                'name': 'Creative_Type',
                'type': 'STRING'
              },
              {
                'name': 'Impressions',
                'type': 'INTEGER'
              },
              {
                'name': 'Clicks',
                'type': 'INTEGER'
              },
              {
                'name': 'Conversions',
                'type': 'FLOAT'
              },
              {
                'name': 'Cost',
                'type': 'FLOAT'
              },
              {
                'name': 'Revenue',
                'type': 'FLOAT'
              },
              {
                'name': 'AV_Impressions_Measurable',
                'type': 'INTEGER'
              },
              {
                'name': 'AV_Impressions_Viewable',
                'type': 'INTEGER'
              }
            ]
          }
        }
      }
    },
    {
      'bigquery': {
        'description': 'The logic query for Deal Finder, transforms report into view used by datastudio.',
        'hour': [
          4
        ],
        'auth': {
          'field': {
            'name': 'auth_write',
            'kind': 'authentication',
            'order': 1,
            'default': 'service',
            'description': 'Credentials used for writing data.'
          }
        },
        'from': {
          'query': "SELECT Partner, Partner_ID, Advertiser, Advertiser_ID, Site, Site_ID, Inventory, Inventory_Type, Creative_Type, Creative_Size, Always_On, Deal_Impressions, Open_Impressions, Rank_Impressions, Deal_Clicks, Open_Clicks, Rank_Clicks, Deal_Conversions, Open_Conversions, Rank_Conversions, Deal_Impressions_Viewable, Open_Impressions_Viewable, Rank_Impressions_Viewable, Deal_Impressions_Measurable, Open_Impressions_Measurable, Rank_Impressions_Measurable, Deal_Cost, Open_Cost, Rank_Cost, FROM ( SELECT FIRST(Partner) AS Partner, FIRST(Partner_ID) AS Partner_ID, FIRST(Advertiser) AS Advertiser, Advertiser_ID, First(Site) AS Site, Site_ID, Inventory, Inventory_Type, Creative_Type, Creative_Width + ' x ' + Creative_Height AS Creative_Size, IF (LEFT(Inventory, 5) == 'AO - ', True, False) AS Always_On, SUM(Deal_Impressions) AS Deal_Impressions, SUM(Open_Impressions) AS Open_Impressions, SUM(Open_Impressions) + SUM(Deal_Impressions) AS Total_Impressions, ROW_NUMBER() OVER (PARTITION BY Advertiser_ID ORDER BY Total_Impressions DESC) AS Rank_Impressions, SUM(Deal_Clicks) AS Deal_Clicks, SUM(Open_Clicks) AS Open_Clicks, SUM(Open_Clicks) + SUM(Deal_Clicks) AS Total_Clicks, ROW_NUMBER() OVER (PARTITION BY Advertiser_ID ORDER BY Total_Clicks DESC) AS Rank_Clicks, SUM(Deal_Conversions) AS Deal_Conversions, SUM(Open_Conversions) AS Open_Conversions, SUM(Open_Conversions) + SUM(Deal_Conversions) AS Total_Conversions, ROW_NUMBER() OVER (PARTITION BY Advertiser_ID ORDER BY Total_Conversions DESC) AS Rank_Conversions, SUM(Deal_Cost) AS Deal_Cost, SUM(Open_Cost) AS Open_Cost, SUM(Open_Cost) + SUM(Deal_Cost) AS Total_Cost, RANK() OVER (PARTITION BY Advertiser_ID ORDER BY Total_Cost DESC) AS Rank_Cost, SUM(Deal_Impressions_Viewable) AS Deal_Impressions_Viewable, SUM(Open_Impressions_Viewable) AS Open_Impressions_Viewable, SUM(Open_Impressions_Viewable) + SUM(Deal_Impressions_Viewable) AS Total_Impressions_Viewable, ROW_NUMBER() OVER (PARTITION BY Advertiser_ID ORDER BY Total_Impressions_Viewable DESC) AS Rank_Impressions_Viewable, SUM(Deal_Impressions_Measurable) AS Deal_Impressions_Measurable, SUM(Open_Impressions_Measurable) AS Open_Impressions_Measurable, SUM(Open_Impressions_Measurable) + SUM(Deal_Impressions_Measurable) AS Total_Impressions_Measurable, ROW_NUMBER() OVER (PARTITION BY Advertiser_ID ORDER BY Total_Impressions_Measurable DESC) AS Rank_Impressions_Measurable, FROM ( SELECT Partner, Partner_ID, Advertiser, Advertiser_ID, Site, Site_ID, Inventory, Inventory_Type, Creative_Type, Creative_Width, Creative_Height, IF(Inventory_ID IS NULL, Impressions, 0) AS Open_Impressions, IF(Inventory_ID IS NULL, 0, Impressions) AS Deal_Impressions, IF(Inventory_ID IS NULL, Clicks, 0) AS Open_Clicks, IF(Inventory_ID IS NULL, 0, Clicks) AS Deal_Clicks, IF(Inventory_ID IS NULL, Conversions, 0) AS Open_Conversions, IF(Inventory_ID IS NULL, 0, Conversions) AS Deal_Conversions, IF(Inventory_ID IS NULL, Cost, 0) AS Open_Cost, IF(Inventory_ID IS NULL, 0, Cost) AS Deal_Cost, IF(Inventory_ID IS NULL, AV_Impressions_Viewable, 0) AS Open_Impressions_Viewable, IF(Inventory_ID IS NULL, 0, AV_Impressions_Viewable) AS Deal_Impressions_Viewable, IF(Inventory_ID IS NULL, AV_Impressions_Measurable, 0) AS Open_Impressions_Measurable, IF(Inventory_ID IS NULL, 0, AV_Impressions_Measurable) AS Deal_Impressions_Measurable, FROM [[PARAMETER].Deal_Finder_DV360_Report] OMIT RECORD IF Site == 'Low volume inventory') GROUP By Advertiser_ID, Site_ID, Inventory, Inventory_Type, Creative_Type, Creative_Size, Always_On) WHERE Rank_Impressions < 100 OR Rank_Clicks < 100 OR Rank_Conversions < 100 OR Rank_Cost < 100;",
          'parameters': [
            {
              'field': {
                'name': 'recipe_slug',
                'kind': 'string',
                'description': 'Place where tables will be written in BigQuery.'
              }
            }
          ]
        },
        'to': {
          'dataset': {
            'field': {
              'name': 'recipe_slug',
              'kind': 'string',
              'description': 'Place where tables will be written in BigQuery.'
            }
          },
          'view': 'Deal_Finder_Dashboard'
        }
      }
    }
  ]
}

DAG_FACTORY = DAG_Factory('deal_finder', RECIPE, INPUTS)
DAG = DAG_FACTORY.generate()

if __name__ == "__main__":
  DAG_FACTORY.print_commandline()



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

from starthinker.util.project import project
from starthinker.util.storage import bucket_create, bucket_access


@project.from_parameters
def bucket():
  if project.verbose:
    print('BUCKET', project.task['bucket'])

  # create bucket
  bucket_create(project.task['auth'], project.id, project.task['bucket'])
  bucket_access(
      project.task['auth'],
      project.id,
      project.task['bucket'],
      emails=project.task.get('emails', []),
      groups=project.task.get('groups', []))


if __name__ == '__main__':
  bucket()



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

'''
--------------------------------------------------------------

Before running this Airflow module...

  Install StarThinker in cloud composer ( recommended ):

    From Release: pip install starthinker
    From Open Source: pip install git+https://github.com/google/starthinker

  Or push local code to the cloud composer plugins directory ( if pushing local code changes ):

    source install/deploy.sh
    4) Composer Menu
    l) Install All

--------------------------------------------------------------

  If any recipe task has "auth" set to "user" add user credentials:

    1. Ensure an RECIPE['setup']['auth']['user'] = [User Credentials JSON]

  OR

    1. Visit Airflow UI > Admin > Connections.
    2. Add an Entry called "starthinker_user", fill in the following fields. Last step paste JSON from authentication.
      - Conn Type: Google Cloud Platform
      - Project: Get from https://github.com/google/starthinker/blob/master/tutorials/cloud_project.md
      - Keyfile JSON: Get from: https://github.com/google/starthinker/blob/master/tutorials/deploy_commandline.md#optional-setup-user-credentials

--------------------------------------------------------------

  If any recipe task has "auth" set to "service" add service credentials:

    1. Ensure an RECIPE['setup']['auth']['service'] = [Service Credentials JSON]

  OR

    1. Visit Airflow UI > Admin > Connections.
    2. Add an Entry called "starthinker_service", fill in the following fields. Last step paste JSON from authentication.
      - Conn Type: Google Cloud Platform
      - Project: Get from https://github.com/google/starthinker/blob/master/tutorials/cloud_project.md
      - Keyfile JSON: Get from: https://github.com/google/starthinker/blob/master/tutorials/cloud_service.md

--------------------------------------------------------------

Trends Places To BigQuery Via Query

Move using a WOEID query.

  - Provide <a href='https://apps.twitter.com/' target='_blank'>Twitter credentials</a>.
  - Provide BigQuery WOEID source query.
  - Specify BigQuery dataset and table to write API call results to.
  - Writes: WOEID, Name, Url, Promoted_Content, Query, Tweet_Volume
  - Note Twitter API is rate limited to 15 requests per 15 minutes. So keep WOEID lists short.

--------------------------------------------------------------

This StarThinker DAG can be extended with any additional tasks from the following sources:
  - https://google.github.io/starthinker/
  - https://github.com/google/starthinker/tree/master/dags

'''

from starthinker.airflow.factory import DAG_Factory

INPUTS = {
  'auth_write': 'service',  # Credentials used for writing data.
  'secret': '',
  'key': '',
  'places_dataset': '',
  'places_query': '',
  'places_legacy': False,
  'destination_dataset': '',
  'destination_table': '',
}

RECIPE = {
  'tasks': [
    {
      'twitter': {
        'auth': {
          'field': {
            'name': 'auth_write',
            'kind': 'authentication',
            'order': 1,
            'default': 'service',
            'description': 'Credentials used for writing data.'
          }
        },
        'secret': {
          'field': {
            'name': 'secret',
            'kind': 'string',
            'order': 1,
            'default': ''
          }
        },
        'key': {
          'field': {
            'name': 'key',
            'kind': 'string',
            'order': 2,
            'default': ''
          }
        },
        'trends': {
          'places': {
            'single_cell': True,
            'bigquery': {
              'dataset': {
                'field': {
                  'name': 'places_dataset',
                  'kind': 'string',
                  'order': 3,
                  'default': ''
                }
              },
              'query': {
                'field': {
                  'name': 'places_query',
                  'kind': 'string',
                  'order': 4,
                  'default': ''
                }
              },
              'legacy': {
                'field': {
                  'name': 'places_legacy',
                  'kind': 'boolean',
                  'order': 5,
                  'default': False
                }
              }
            }
          }
        },
        'out': {
          'bigquery': {
            'dataset': {
              'field': {
                'name': 'destination_dataset',
                'kind': 'string',
                'order': 6,
                'default': ''
              }
            },
            'table': {
              'field': {
                'name': 'destination_table',
                'kind': 'string',
                'order': 7,
                'default': ''
              }
            }
          }
        }
      }
    }
  ]
}

DAG_FACTORY = DAG_Factory('trends_places_to_bigquery_via_query', RECIPE, INPUTS)
DAG = DAG_FACTORY.generate()

if __name__ == "__main__":
  DAG_FACTORY.print_commandline()



from starthinker.util.project import project
from starthinker.util.dynamite import dynamite_write


@project.from_parameters
def dynamite():
  if project.verbose:
    print('DYNAMITE')
  dynamite_write(project.task['room'], project.task['key'],
                 project.task['token'], project.task['message'])


if __name__ == '__main__':
  dynamite()



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

'''
--------------------------------------------------------------

Before running this Airflow module...

  Install StarThinker in cloud composer ( recommended ):

    From Release: pip install starthinker
    From Open Source: pip install git+https://github.com/google/starthinker

  Or push local code to the cloud composer plugins directory ( if pushing local code changes ):

    source install/deploy.sh
    4) Composer Menu
    l) Install All

--------------------------------------------------------------

  If any recipe task has "auth" set to "user" add user credentials:

    1. Ensure an RECIPE['setup']['auth']['user'] = [User Credentials JSON]

  OR

    1. Visit Airflow UI > Admin > Connections.
    2. Add an Entry called "starthinker_user", fill in the following fields. Last step paste JSON from authentication.
      - Conn Type: Google Cloud Platform
      - Project: Get from https://github.com/google/starthinker/blob/master/tutorials/cloud_project.md
      - Keyfile JSON: Get from: https://github.com/google/starthinker/blob/master/tutorials/deploy_commandline.md#optional-setup-user-credentials

--------------------------------------------------------------

  If any recipe task has "auth" set to "service" add service credentials:

    1. Ensure an RECIPE['setup']['auth']['service'] = [Service Credentials JSON]

  OR

    1. Visit Airflow UI > Admin > Connections.
    2. Add an Entry called "starthinker_service", fill in the following fields. Last step paste JSON from authentication.
      - Conn Type: Google Cloud Platform
      - Project: Get from https://github.com/google/starthinker/blob/master/tutorials/cloud_project.md
      - Keyfile JSON: Get from: https://github.com/google/starthinker/blob/master/tutorials/cloud_service.md

--------------------------------------------------------------

Twitter Targeting

Adjusts line item settings based on Twitter hashtags and locations specified in a sheet.

  - Click <b>Run Now</b> and a sheet called <b>Twitter Targeting UNDEFINED</b> will be generated with a tab called <b>Twitter Triggers</b>.
  - Follow instructions on the sheets tab to provide triggers and lineitems.
  - Click <b>Run Now</b> again, trends are downloaded and triggered.
  - Or give these intructions to the client.

--------------------------------------------------------------

This StarThinker DAG can be extended with any additional tasks from the following sources:
  - https://google.github.io/starthinker/
  - https://github.com/google/starthinker/tree/master/dags

'''

from starthinker.airflow.factory import DAG_Factory

INPUTS = {
  'recipe_slug': '',  # Place where tables will be created in BigQuery.
  'auth_read': 'user',  # Credentials used for reading data.
  'recipe_project': '',  # Place where tables will be created in BigQuery.
  'auth_write': 'service',  # Credentials used for writing data.
  'recipe_name': '',  # Name of sheet where Line Item settings will be read from.
  'twitter_secret': '',  # Twitter API secret token.
  'twitter_key': '',  # Twitter API key token.
}

RECIPE = {
  'setup': {
    'day': [
      'Mon',
      'Tue',
      'Wed',
      'Thu',
      'Fri',
      'Sat',
      'Sun'
    ],
    'hour': [
      2,
      4,
      6,
      8,
      10,
      12,
      14,
      16,
      18,
      20,
      22,
      24
    ]
  },
  'tasks': [
    {
      'dataset': {
        'description': 'Create a dataset where data will be combined and transfored for upload.',
        'auth': {
          'field': {
            'name': 'auth_write',
            'kind': 'authentication',
            'order': 1,
            'default': 'service',
            'description': 'Credentials used for writing data.'
          }
        },
        'dataset': {
          'field': {
            'name': 'recipe_slug',
            'kind': 'string',
            'order': 1,
            'description': 'Place where tables will be created in BigQuery.'
          }
        }
      }
    },
    {
      'sheets': {
        'description': 'Read mapping of hash tags to line item toggles from sheets.',
        'auth': {
          'field': {
            'name': 'auth_read',
            'kind': 'authentication',
            'order': 0,
            'default': 'user',
            'description': 'Credentials used for reading data.'
          }
        },
        'template': {
          'sheet': 'https://docs.google.com/spreadsheets/d/1iYCGa2NKOZiL2mdT4yiDfV_SWV9C7SUosXdIr4NAEXE/edit?usp=sharing',
          'tab': 'Twitter Triggers'
        },
        'sheet': {
          'field': {
            'name': 'recipe_name',
            'kind': 'string',
            'prefix': 'Twitter Targeting For ',
            'order': 2,
            'description': 'Name of sheet where Line Item settings will be read from.',
            'default': ''
          }
        },
        'tab': 'Twitter Triggers',
        'range': 'A8:T',
        'out': {
          'bigquery': {
            'dataset': {
              'field': {
                'name': 'recipe_slug',
                'kind': 'string',
                'description': 'Place where tables will be created in BigQuery.'
              }
            },
            'table': 'Twitter_Triggers',
            'schema': [
              {
                'name': 'Location',
                'type': 'STRING',
                'mode': 'REQUIRED'
              },
              {
                'name': 'WOEID',
                'type': 'INTEGER',
                'mode': 'REQUIRED'
              },
              {
                'name': 'Hashtag',
                'type': 'STRING',
                'mode': 'REQUIRED'
              },
              {
                'name': 'Line_Item_Id',
                'type': 'INTEGER',
                'mode': 'REQUIRED'
              },
              {
                'name': 'Line_Item_Name',
                'type': 'STRING',
                'mode': 'NULLABLE'
              },
              {
                'name': 'Line_Item_Status',
                'type': 'STRING',
                'mode': 'NULLABLE'
              },
              {
                'name': 'Line_Item_Start_Date',
                'type': 'STRING',
                'mode': 'NULLABLE'
              },
              {
                'name': 'Line_Item_End_Date',
                'type': 'STRING',
                'mode': 'NULLABLE'
              },
              {
                'name': 'Line_Item_Budget_Type',
                'type': 'STRING',
                'mode': 'NULLABLE'
              },
              {
                'name': 'Line_Item_Budget_Amount',
                'type': 'STRING',
                'mode': 'NULLABLE'
              },
              {
                'name': 'Line_Item_Pacing',
                'type': 'STRING',
                'mode': 'NULLABLE'
              },
              {
                'name': 'Line_Item_Pacing_Rate',
                'type': 'STRING',
                'mode': 'NULLABLE'
              },
              {
                'name': 'Line_Item_Pacing_Amount',
                'type': 'STRING',
                'mode': 'NULLABLE'
              },
              {
                'name': 'Line_Item_Frequency_Enabled',
                'type': 'STRING',
                'mode': 'NULLABLE'
              },
              {
                'name': 'Line_Item_Frequency_Exposures',
                'type': 'STRING',
                'mode': 'NULLABLE'
              },
              {
                'name': 'Line_Item_Frequency_Period',
                'type': 'STRING',
                'mode': 'NULLABLE'
              },
              {
                'name': 'Line_Item_Frequency_Amount',
                'type': 'STRING',
                'mode': 'NULLABLE'
              },
              {
                'name': 'Bid_Price',
                'type': 'STRING',
                'mode': 'NULLABLE'
              },
              {
                'name': 'Partner_Revenue_Model',
                'type': 'STRING',
                'mode': 'NULLABLE'
              },
              {
                'name': 'Partner_Revenue_Amount',
                'type': 'STRING',
                'mode': 'NULLABLE'
              }
            ]
          }
        }
      }
    },
    {
      'twitter': {
        'description': 'Read trends from Twitter and place into BigQuery.',
        'auth': {
          'field': {
            'name': 'auth_write',
            'kind': 'authentication',
            'order': 1,
            'default': 'service',
            'description': 'Credentials used for writing data.'
          }
        },
        'secret': {
          'field': {
            'name': 'twitter_secret',
            'kind': 'string',
            'order': 3,
            'default': '',
            'description': 'Twitter API secret token.'
          }
        },
        'key': {
          'field': {
            'name': 'twitter_key',
            'kind': 'string',
            'order': 4,
            'default': '',
            'description': 'Twitter API key token.'
          }
        },
        'trends': {
          'places': {
            'single_cell': True,
            'bigquery': {
              'dataset': {
                'field': {
                  'name': 'recipe_slug',
                  'kind': 'string',
                  'description': 'Place where tables will be created in BigQuery.'
                }
              },
              'query': 'SELECT WOEID FROM {dataset}.Twitter_Triggers',
              'legacy': False,
              'parameters': {
                'dataset': {
                  'field': {
                    'name': 'recipe_slug',
                    'kind': 'string',
                    'description': 'Place where tables will be created in BigQuery.'
                  }
                }
              }
            }
          }
        },
        'out': {
          'bigquery': {
            'dataset': {
              'field': {
                'name': 'recipe_slug',
                'kind': 'string',
                'description': 'Place where tables will be created in BigQuery.'
              }
            },
            'table': 'Twitter_Trends_Place'
          }
        }
      }
    },
    {
      'lineitem': {
        'description': 'Read current lineitem settings from DV360 into BigQuery, so it can be joined with Twitter analysis.',
        'auth': {
          'field': {
            'name': 'auth_read',
            'kind': 'authentication',
            'order': 0,
            'default': 'user',
            'description': 'Credentials used for reading data.'
          }
        },
        'read': {
          'line_items': {
            'single_cell': True,
            'bigquery': {
              'dataset': {
                'field': {
                  'name': 'recipe_slug',
                  'kind': 'string',
                  'description': 'Place where tables will be created in BigQuery.'
                }
              },
              'query': 'SELECT Line_Item_Id FROM {dataset}.Twitter_Triggers',
              'parameters': {
                'dataset': {
                  'field': {
                    'name': 'recipe_slug',
                    'kind': 'string',
                    'description': 'Place where tables will be created in BigQuery.'
                  }
                }
              }
            }
          },
          'out': {
            'bigquery': {
              'dataset': {
                'field': {
                  'name': 'recipe_slug',
                  'kind': 'string',
                  'description': 'Place where tables will be created in BigQuery.'
                }
              },
              'table': 'LineItem_Reads'
            }
          }
        }
      }
    },
    {
      'bigquery': {
        'description': 'Get all triggered lineitmes from sheet, if they have a keyword match in twitter, take the triger values, else take the default values (default>trigger).  Take all non-null values from trigger and overlay over current DV360 values. Will be used to upload to DV360.',
        'auth': {
          'field': {
            'name': 'auth_write',
            'kind': 'authentication',
            'order': 1,
            'default': 'service',
            'description': 'Credentials used for writing data.'
          }
        },
        'from': {
          'query': "SELECT o.Line_Item_Id AS Line_Item_Id, o.Partner_Name AS Partner_Name, o.Partner_Id AS Partner_Id, o.Advertiser_Name AS Advertiser_Name, o.IO_Name AS IO_Name, IFNULL(t.Line_Item_Name, o.Line_Item_Name) AS Line_Item_Name, o.Line_Item_Timestamp AS Line_Item_Timestamp , IFNULL(t.Line_Item_Status, o.Line_Item_Status) AS Line_Item_Status, o.IO_Start_Date AS IO_Start_Date, o.IO_End_Date AS IO_End_Date, o.IO_Budget_Type AS IO_Budget_Type, o.IO_Budget_Amount AS IO_Budget_Amount, o.IO_Pacing AS IO_Pacing, o.IO_Pacing_Rate AS IO_Pacing_Rate, o.IO_Pacing_Amount AS IO_Pacing_Amount, IFNULL(t.Line_Item_Start_Date, o.Line_Item_Start_Date) AS Line_Item_Start_Date, IFNULL(t.Line_Item_End_Date, o.Line_Item_End_Date) AS Line_Item_End_Date, IFNULL(t.Line_Item_Budget_Type, o.Line_Item_Budget_Type) AS Line_Item_Budget_Type, IFNULL(t.Line_Item_Budget_Amount, o.Line_Item_Budget_Amount) AS Line_Item_Budget_Amount, IFNULL(t.Line_Item_Pacing, o.Line_Item_Pacing) AS Line_Item_Pacing, IFNULL(t.Line_Item_Pacing_Rate, o.Line_Item_Pacing_Rate) AS Line_Item_Pacing_Rate, IFNULL(t.Line_Item_Pacing_Amount, o.Line_Item_Pacing_Amount) AS Line_Item_Pacing_Amount, IFNULL(t.Line_Item_Frequency_Enabled, o.Line_Item_Frequency_Enabled) AS Line_Item_Frequency_Enabled, IFNULL(t.Line_Item_Frequency_Exposures, o.Line_Item_Frequency_Exposures) AS Line_Item_Frequency_Exposures, IFNULL(t.Line_Item_Frequency_Period, o.Line_Item_Frequency_Period) AS Line_Item_Frequency_Period, IFNULL(t.Line_Item_Frequency_Amount, o.Line_Item_Frequency_Amount) AS Line_Item_Frequency_Amount, IFNULL(t.Bid_Price, o.Bid_Price) AS Bid_Price, IFNULL(t.Partner_Revenue_Model, o.Partner_Revenue_Model) AS Partner_Revenue_Model, IFNULL(t.Partner_Revenue_Amount, o.Partner_Revenue_Amount) AS Partner_Revenue_Amount, o.Current_Audience_Targeting_Ids AS Current_Audience_Targeting_Ids , o.Current_Audience_Targeting_Names AS Current_Audience_Targeting_Names FROM `{project}.{dataset}.LineItem_Reads` AS o LEFT JOIN ( SELECT Line_Item_Id, ANY_VALUE(SPLIT(Line_Item_Name, '>')[SAFE_OFFSET(IF(Triggered, 1, 0))]) AS Line_Item_Name, ANY_VALUE(SPLIT(Line_Item_Status, '>')[SAFE_OFFSET(IF(Triggered, 1, 0))]) AS Line_Item_Status, ANY_VALUE(SPLIT(Line_Item_Start_Date, '>')[SAFE_OFFSET(IF(Triggered, 1, 0))]) AS Line_Item_Start_Date, ANY_VALUE(SPLIT(Line_Item_End_Date, '>')[SAFE_OFFSET(IF(Triggered, 1, 0))]) AS Line_Item_End_Date, ANY_VALUE(SPLIT(Line_Item_Budget_Type, '>')[SAFE_OFFSET(IF(Triggered, 1, 0))]) AS Line_Item_Budget_Type, ANY_VALUE(CAST(SPLIT(Line_Item_Budget_Amount, '>')[SAFE_OFFSET(IF(Triggered, 1, 0))] AS FLOAT64)) AS Line_Item_Budget_Amount, ANY_VALUE(SPLIT(Line_Item_Pacing, '>')[SAFE_OFFSET(IF(Triggered, 1, 0))]) AS Line_Item_Pacing, ANY_VALUE(SPLIT(Line_Item_Pacing_Rate, '>')[SAFE_OFFSET(IF(Triggered, 1, 0))]) AS Line_Item_Pacing_Rate, ANY_VALUE(CAST(SPLIT(Line_Item_Pacing_Amount, '>')[SAFE_OFFSET(IF(Triggered, 1, 0))] AS FLOAT64)) AS Line_Item_Pacing_Amount, ANY_VALUE(CAST(SPLIT(Line_Item_Frequency_Enabled, '>')[SAFE_OFFSET(IF(Triggered, 1, 0))] AS BOOL)) AS Line_Item_Frequency_Enabled, ANY_VALUE(SPLIT(Line_Item_Frequency_Exposures, '>')[SAFE_OFFSET(IF(Triggered, 1, 0))]) AS Line_Item_Frequency_Exposures, ANY_VALUE(SPLIT(Line_Item_Frequency_Period, '>')[SAFE_OFFSET(IF(Triggered, 1, 0))]) AS Line_Item_Frequency_Period, ANY_VALUE(CAST(SPLIT(Line_Item_Frequency_Amount, '>')[SAFE_OFFSET(IF(Triggered, 1, 0))] AS INT64)) AS Line_Item_Frequency_Amount, ANY_VALUE(CAST(SPLIT(Bid_Price, '>')[SAFE_OFFSET(IF(Triggered, 1, 0))] AS FLOAT64)) AS Bid_Price, ANY_VALUE(SPLIT(Partner_Revenue_Model, '>')[SAFE_OFFSET(IF(Triggered, 1, 0))]) AS Partner_Revenue_Model, ANY_VALUE(CAST(SPLIT(Partner_Revenue_Amount, '>')[SAFE_OFFSET(IF(Triggered, 1, 0))] AS FLOAT64)) AS Partner_Revenue_Amount FROM ( SELECT WOEID, Hashtag, Line_Item_Id, Line_Item_Name, Line_Item_Status, Line_Item_Start_Date, Line_Item_End_Date, Line_Item_Budget_Type, Line_Item_Budget_Amount, Line_Item_Pacing, Line_Item_Pacing_Rate, Line_Item_Pacing_Amount, Line_Item_Frequency_Enabled, Line_Item_Frequency_Exposures, Line_Item_Frequency_Period, Line_Item_Frequency_Amount, Bid_Price, Partner_Revenue_Model, Partner_Revenue_Amount, CONCAT(CAST(WOEID AS STRING), LOWER(Hashtag)) IN (SELECT CONCAT(CAST(WOEID AS STRING), LOWER(REPLACE(name, '#', ''))) FROM `{project}.{dataset}.Twitter_Trends_Place` GROUP BY 1) AS Triggered FROM `{project}.{dataset}.Twitter_Triggers`) GROUP BY 1) AS t ON o.Line_Item_Id=t.Line_Item_Id;",
          'parameters': {
            'project': {
              'field': {
                'name': 'recipe_project',
                'kind': 'string',
                'description': 'Place where tables will be created in BigQuery.'
              }
            },
            'dataset': {
              'field': {
                'name': 'recipe_slug',
                'kind': 'string',
                'description': 'Place where tables will be created in BigQuery.'
              }
            }
          },
          'legacy': False
        },
        'to': {
          'dataset': {
            'field': {
              'name': 'recipe_slug',
              'kind': 'string',
              'description': 'Place where tables will be created in BigQuery.'
            }
          },
          'view': 'LineItem_Writes'
        }
      }
    },
    {
      'lineitem': {
        'description': 'Write lineitem settings to DV360 after transformation.',
        'auth': {
          'field': {
            'name': 'auth_read',
            'kind': 'authentication',
            'order': 0,
            'default': 'user',
            'description': 'Credentials used for reading data.'
          }
        },
        'write': {
          'dry_run': False,
          'bigquery': {
            'dataset': {
              'field': {
                'name': 'recipe_slug',
                'kind': 'string',
                'description': 'Place where tables will be created in BigQuery.'
              }
            },
            'query': 'Select * FROM {dataset}.LineItem_Writes',
            'parameters': {
              'dataset': {
                'field': {
                  'name': 'recipe_slug',
                  'kind': 'string',
                  'description': 'Place where tables will be created in BigQuery.'
                }
              }
            },
            'legacy': False
          }
        }
      }
    }
  ]
}

DAG_FACTORY = DAG_Factory('twitter', RECIPE, INPUTS)
DAG = DAG_FACTORY.generate()

if __name__ == "__main__":
  DAG_FACTORY.print_commandline()



# -*- coding: utf-8 -*-

###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

from django.conf import settings
from django.contrib import messages
from django.contrib.auth import login as django_login, logout as django_logout
from django.http import HttpResponseRedirect

from starthinker.util.auth.wrapper import CredentialsFlowWrapper
from starthinker_ui.account.models import Account


def oauth_callback(request):

  # get the credentials from the Google redirect
  flow = CredentialsFlowWrapper(
      settings.UI_CLIENT, redirect_uri=settings.CONST_URL + '/oauth_callback/')
  flow.code_verifier = request.session.get('code_verifier')
  flow.fetch_token(code=request.GET['code'])

  # get or create the account
  account = Account.objects.get_or_create_user(flow.credentials)

  # log the account in ( set cookie )
  django_login(request, account, backend=settings.AUTHENTICATION_BACKENDS[0])

  messages.success(request, 'Welcome To StarThinker')

  return HttpResponseRedirect('/')


def logout(request):
  django_logout(request)
  messages.success(request, 'You Are Logged Out')
  return HttpResponseRedirect('/')



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

'''
--------------------------------------------------------------

Before running this Airflow module...

  Install StarThinker in cloud composer from open source:

    pip install git+https://github.com/google/starthinker

  Or push local code to the cloud composer plugins directory:

    source install/deploy.sh
    4) Composer Menu
    l) Install All

--------------------------------------------------------------

DV360 API Patch From BigQuery

Patch DV360 API endpoints.

Specify the name of the dataset and table.
Rows will be read and applied as a patch to DV360.

'''

from starthinker_airflow.factory import DAG_Factory

# Add the following credentials to your Airflow configuration.
USER_CONN_ID = "starthinker_user" # The connection to use for user authentication.
GCP_CONN_ID = "starthinker_service" # The connection to use for service authentication.

INPUTS = {
  'auth_write': 'user',  # Credentials used for writing data.
  'patch': '',
  'auth_read': 'service',  # Credentials used for reading data.
  'dataset': '',  # Google BigQuery dataset to create tables in.
  'table': '',  # Google BigQuery dataset to create tables in.
}

TASKS = [
  {
    'dv360_api': {
      'auth': {
        'field': {
          'name': 'auth_write',
          'kind': 'authentication',
          'order': 0,
          'default': 'user',
          'description': 'Credentials used for writing data.'
        }
      },
      'patch': {
        'field': {
          'name': 'patch',
          'kind': 'choice',
          'choices': [
            'advertisers',
            'advertisers.campaigns',
            'advertisers.channels',
            'advertisers.channels.sites',
            'advertisers.creatives',
            'advertisers.insertionOrders',
            'advertisers.lineItems',
            'advertisers.locationLists',
            'advertisers.locationLists.assignedLocations',
            'advertisers.negativeKeywordLists',
            'advertisers.negativeKeywordLists.negativeKeywords',
            'floodlightGroups',
            'inventorySourceGroups',
            'partners.channels',
            'users'
          ],
          'default': ''
        }
      },
      'bigquery': {
        'auth': {
          'field': {
            'name': 'auth_read',
            'kind': 'authentication',
            'order': 1,
            'default': 'service',
            'description': 'Credentials used for reading data.'
          }
        },
        'dataset': {
          'field': {
            'name': 'dataset',
            'kind': 'string',
            'order': 2,
            'default': '',
            'description': 'Google BigQuery dataset to create tables in.'
          }
        },
        'table': {
          'field': {
            'name': 'table',
            'kind': 'string',
            'order': 3,
            'default': '',
            'description': 'Google BigQuery dataset to create tables in.'
          }
        },
        'as_object': True
      }
    }
  }
]

DAG_FACTORY = DAG_Factory('dv360_api_patch_from_bigquery', { 'tasks':TASKS }, INPUTS)
DAG_FACTORY.apply_credentails(USER_CONN_ID, GCP_CONN_ID)
DAG = DAG_FACTORY.execute()

if __name__ == "__main__":
  DAG_FACTORY.print_commandline()



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

from starthinker.script.parse import dict_to_string
from starthinker.script.parse import fields_to_string
from starthinker.script.parse import json_get_fields
from starthinker.script.parse import json_set_auths
from starthinker.util.colab import Colab


def script_to_colab(name, description, instructions, tasks, parameters={}):
  colab = Colab(name)

  colab.header('1. Install Dependencies')
  colab.paragraph(
      'First install the libraries needed to execute recipes, this only needs to be done once, then click play.'
  )
  colab.code('!pip install git+https://github.com/google/starthinker')

  colab.header('2. Get Cloud Project ID')
  colab.paragraph(
      'To run this recipe [requires a Google Cloud Project](https://github.com/google/starthinker/blob/master/tutorials/cloud_project.md), this only needs to be done once, then click play.'
  )
  #colab.image('Client Project ID', 'https://github.com/google/starthinker/raw/master/tutorials/images/cloud_project.png')
  colab.code('CLOUD_PROJECT = \'PASTE PROJECT ID HERE\'')
  colab.code('\nprint("Cloud Project Set To: %s" % CLOUD_PROJECT)')

  colab.header('3. Get Client Credentials')
  #colab.image('Client Credentials', 'https://github.com/google/starthinker/raw/master/tutorials/images/cloud_client_installed.png')
  colab.paragraph(
      'To read and write to various endpoints requires [downloading client credentials](https://github.com/google/starthinker/blob/master/tutorials/cloud_client_installed.md), this only needs to be done once, then click play.'
  )
  colab.code('CLIENT_CREDENTIALS = \'PASTE CREDENTIALS HERE\'')
  colab.code('\nprint("Client Credentials Set To: %s" % CLIENT_CREDENTIALS)')

  fields = json_get_fields(tasks)
  if fields:
    colab.header('4. Enter %s Parameters' % name)
    colab.paragraph(description)
    colab.list(instructions)
    colab.paragraph(
        'Modify the values below for your use case, can be done multiple times, then click play.'
    )
    colab.code('FIELDS = %s' % fields_to_string(fields, parameters))
    colab.code('\nprint("Parameters Set To: %s" % FIELDS)')

  colab.header('%d. Execute %s' % (5 if fields else 4, name))
  colab.paragraph(
      'This does NOT need to be modified unles you are changing the recipe, click play.'
  )

  colab.code('from starthinker.util.project import project')
  colab.code('from starthinker.script.parse import json_set_fields')
  colab.code('')
  colab.code("USER_CREDENTIALS = '/content/user.json'")
  colab.code('')
  colab.code('TASKS = %s' %
             dict_to_string(json_set_auths(tasks, 'user'), skip=('field',)))
  colab.code('')

  if fields:
    colab.code('json_set_fields(TASKS, FIELDS)')
  colab.code('')

  colab.code(
      "project.initialize(_recipe={ 'tasks':TASKS }, _project=CLOUD_PROJECT, _user=USER_CREDENTIALS, _client=CLIENT_CREDENTIALS, _verbose=True, _force=True)"
  )
  colab.code('project.execute(_force=True)')

  return colab.render()



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

import csv
import io
import json

from django import forms
from django.core import validators
from django.core.exceptions import ValidationError

from starthinker_ui.recipe.timezones import TIMEZONES


class CommaSeparatedCharField(forms.CharField):

  def prepare_value(self, value):
    return ', '.join(value or [])

  def clean(self, value):
    try:
      v = [str(v.strip()) for v in value.split(',') if v.strip()]
      json.dumps(v)  # only do a check if valid
      return v
    except:
      raise ValidationError


class CommaSeparatedIntegerField(forms.Field):
  default_error_messages = {
      'invalid': 'Enter comma separated numbers only.',
  }

  def prepare_value(self, value):
    return ', '.join(
        str(v)
        for v in (json.loads(value or '[]') if isinstance(value, str) else (
            value or [])))

  def clean(self, value):
    try:
      v = [int(v.strip()) for v in value.split(',') if v.strip()]
      json.dumps(v)  # only do a check if valid
      return v
    except:
      raise ValidationError


class ListChoiceField(forms.MultipleChoiceField):

  def prepare_value(self, value):
    return json.loads(value or '[]') if isinstance(value, str) else value

  def to_python(self, value):
    return json.dumps(value)

  def validate(self, value):
    try:
      json.loads(value)
      return True
    except:
      return False


class ListChoiceIntegerField(forms.MultipleChoiceField):

  def prepare_value(self, value):
    return json.loads(value or '[]') if isinstance(value, str) else value

  def to_python(self, value):
    return json.dumps([int(v) for v in value])

  def validate(self, value):
    try:
      json.loads(value)
      return True
    except:
      return False


class ListChoiceIntegerField(forms.MultipleChoiceField):

  def prepare_value(self, value):
    return json.loads(value or '[]') if isinstance(value, str) else value

  def to_python(self, value):
    return json.dumps([int(v) for v in value])

  def validate(self, value):
    try:
      json.loads(value)
      return True
    except:
      return False


class JsonField(forms.CharField):
  widget = forms.Textarea

  def prepare_value(self, value):
    if isinstance(value, str):
      if value:
        try:
          return json.dumps(json.loads(value))
        except ValueError:
          return value
      else:
        return None
    else:
      return json.dumps(value)

  def clean(self, value):
    if isinstance(value, str):
      if value:
        try:
          return json.loads(value.strip())
        except ValueError as e:
          raise ValidationError(str(e))
      else:
        return None
    else:
      return value


class TimezoneField(forms.ChoiceField):

  def __init__(self, *args, **kwargs):
    kwargs['choices'] = map(lambda t: (t, t.replace('_', ' ')), TIMEZONES)
    kwargs['initial'] = 'America/Los_Angeles'
    super(TimezoneField, self).__init__(*args, **kwargs)


class SwitchField(forms.ChoiceField):
  widget = forms.CheckboxInput

  def __init__(self, value_false, value_true, *args, **kwargs):
    self.value_true = value_true
    self.value_false = value_false
    kwargs['choices'] = [value_true, value_false]
    super(SwitchField, self).__init__(*args, **kwargs)

  def prepare_value(self, value):
    return value == self.value_true

  def clean(self, value):
    if value == True:
      return self.value_true
    else:
      return self.value_false



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

from starthinker.util.project import project

from starthinker.task.dv_editor.advertiser import advertiser_clear
from starthinker.task.dv_editor.advertiser import advertiser_load
from starthinker.task.dv_editor.audit import audit_clear
from starthinker.task.dv_editor.audit import audit_load
from starthinker.task.dv_editor.bid_strategy import bid_strategy_clear
from starthinker.task.dv_editor.bid_strategy import bid_strategy_load
from starthinker.task.dv_editor.bid_strategy import bid_strategy_patch
from starthinker.task.dv_editor.campaign import campaign_clear
from starthinker.task.dv_editor.campaign import campaign_load
from starthinker.task.dv_editor.creative import creative_clear
from starthinker.task.dv_editor.creative import creative_load
from starthinker.task.dv_editor.frequency_cap import frequency_cap_clear
from starthinker.task.dv_editor.frequency_cap import frequency_cap_load
from starthinker.task.dv_editor.frequency_cap import frequency_cap_patch
from starthinker.task.dv_editor.insertion_order import insertion_order_audit
from starthinker.task.dv_editor.insertion_order import insertion_order_clear
from starthinker.task.dv_editor.insertion_order import insertion_order_insert
from starthinker.task.dv_editor.insertion_order import insertion_order_load
from starthinker.task.dv_editor.insertion_order import insertion_order_patch
from starthinker.task.dv_editor.integration_detail import integration_detail_clear
from starthinker.task.dv_editor.integration_detail import integration_detail_load
from starthinker.task.dv_editor.integration_detail import integration_detail_patch
from starthinker.task.dv_editor.line_item import line_item_audit
from starthinker.task.dv_editor.line_item import line_item_clear
from starthinker.task.dv_editor.line_item import line_item_insert
from starthinker.task.dv_editor.line_item import line_item_load
from starthinker.task.dv_editor.line_item import line_item_patch
from starthinker.task.dv_editor.line_item_map import line_item_map_clear
from starthinker.task.dv_editor.line_item_map import line_item_map_patch
from starthinker.task.dv_editor.pacing import pacing_clear
from starthinker.task.dv_editor.pacing import pacing_load
from starthinker.task.dv_editor.pacing import pacing_patch
from starthinker.task.dv_editor.patch import patch_clear
from starthinker.task.dv_editor.partner import partner_clear
from starthinker.task.dv_editor.partner import partner_load
from starthinker.task.dv_editor.partner_cost import partner_cost_clear
from starthinker.task.dv_editor.partner_cost import partner_cost_load
from starthinker.task.dv_editor.partner_cost import partner_cost_patch
from starthinker.task.dv_editor.segment import segment_clear
from starthinker.task.dv_editor.segment import segment_load
from starthinker.task.dv_editor.segment import segment_patch


@project.from_parameters
def dv_editor():
  print('COMMAND:', project.task['command'])

  if project.task['command'] == 'Load Partners':
    partner_clear()
    partner_load()

  elif project.task['command'] == 'Load Advertisers':
    advertiser_clear()
    advertiser_load()

  elif project.task['command'] == 'Load Campaigns':
    campaign_clear()
    campaign_load()

  elif project.task['command'] == 'Load Creatives':
    creative_clear()
    creative_load()

  elif project.task['command'] == 'Load Insertion Orders':
    insertion_order_clear()
    insertion_order_load()
    segment_clear()
    segment_load()
    pacing_clear()
    pacing_load()
    bid_strategy_clear()
    bid_strategy_load()
    frequency_cap_clear()
    frequency_cap_load()
    partner_cost_clear()
    partner_cost_load()
    integration_detail_clear()
    integration_detail_load()

  elif project.task['command'] == 'Load Line Items':
    line_item_clear()
    line_item_load()
    pacing_clear()
    pacing_load()
    bid_strategy_clear()
    bid_strategy_load()
    frequency_cap_clear()
    frequency_cap_load()
    partner_cost_clear()
    partner_cost_load()
    integration_detail_clear()
    integration_detail_load()

  elif project.task['command'] in ('Preview', 'Patch'):
    audit_clear()
    patch_clear()
    audit_load()

    line_item_insert(commit=project.task['command'] == 'Patch')
    insertion_order_insert(commit=project.task['command'] == 'Patch')

    insertion_order_patch(commit=project.task['command'] == 'Patch')
    line_item_patch(commit=project.task['command'] == 'Patch')
    line_item_map_patch(commit=project.task['command'] == 'Patch')

    segment_patch(commit=project.task['command'] == 'Patch')
    pacing_patch(commit=project.task['command'] == 'Patch')
    bid_strategy_patch(commit=project.task['command'] == 'Patch')
    frequency_cap_patch(commit=project.task['command'] == 'Patch')
    partner_cost_patch(commit=project.task['command'] == 'Patch')
    integration_detail_patch(commit=project.task['command'] == 'Patch')

  elif project.task['command'] == 'Clear Partners':
    partner_clear()

  elif project.task['command'] == 'Clear Advertisers':
    advertiser_clear()

  elif project.task['command'] == 'Clear Campaigns':
    campaign_clear()

  elif project.task['command'] == 'Clear Creatives':
    creative_clear()

  elif project.task['command'] == 'Clear Insertion Orders':
    segment_clear()
    pacing_clear()
    bid_strategy_clear()
    frequency_cap_clear()
    partner_cost_clear()
    integration_detail_clear()

    insertion_order_clear()

  elif project.task['command'] == 'Clear Line Items':
    pacing_clear()
    bid_strategy_clear()
    frequency_cap_clear()
    partner_cost_clear()
    integration_detail_clear()

    line_item_map_clear()
    line_item_clear()

  elif project.task['command'] == 'Clear Preview':
    audit_clear()

  elif project.task['command'] == 'Clear Patch':
    patch_clear()

  elif project.task['command'] == 'Clear All':
    partner_clear()
    advertiser_clear()
    campaign_clear()
    creative_clear()

    segment_clear()
    pacing_clear()
    bid_strategy_clear()
    frequency_cap_clear()
    partner_cost_clear()
    integration_detail_clear()

    insertion_order_clear()
    line_item_map_clear()
    line_item_clear()

    audit_clear()
    patch_clear()

if __name__ == '__main__':
  dv_editor()



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################
"""Simple datastor implentation for storing key/value pairs quickly.

Datastore concepts similar to a database:
- Namespace is like database name.
- Kind is like database table.
- Key is like a primary key.
- Data is like a row or record passed in as a dictionary )

Use cases:

Quickly store a dictionary of data for lookup from different machines.
More specific logging, where key is object ID.

"""

import datetime

from starthinker.util.project import project
from starthinker.util.google_api import API_Datastore


def _datastore_p_to_v(properties):
  """Very simple conversion from datastore types to python types.

  Handles
    - nullValue
    - stringValue
    - integerValue
    - doubleValue
    - booleanValue
    - timestampValue

  Omitted For Simplicity ( add as needed )
    - keyValue
    - entityValue
    - blobValue
    - geoPointValue
    - arrayValue

  """

  v = {}

  for p, d in properties.items():
    for dk, dv in d.items():
      if dk.endswith('Value'):
        if dk == 'nullValue':
          v[p] = dv
        elif dk == 'stringValue':
          v[p] = dv
        elif dk == 'integerValue':
          v[p] = int(dv)
        elif dk == 'doubleValue':
          v[p] = float(dv)
        elif dk == 'booleanValue':
          v[p] = bool(dv)
        elif dk == 'timestampValue':
          v[p] = datetime.datetime.strptime(dv + 'UTC',
                                            '%Y-%m-%dT%H:%M:%S.%fZ%Z')
        else:
          raise Exception(
              'No mapping from python to datastore for type of: %s' % k)

  return v


def _datastore_path(path):
  return '.'.join([p['name'] for p in path])


def _datastore_v_to_p(values):
  """Very simple conversion from datastore types to python types.

  Handles
    - nullValue
    - stringValue
    - integerValue
    - doubleValue
    - booleanValue
    - timestampValue

  Omitted For Simplicity ( add as needed )
    - keyValue
    - entityValue
    - blobValue
    - geoPointValue
    - arrayValue

  """

  p = {}

  for k, v in values.items():
    if v is None:
      p[k] = {'nullValue': v}
    elif isinstance(v, str):
      p[k] = {'stringValue': v, 'excludeFromIndexes': True}
    elif isinstance(v, int):
      p[k] = {'integerValue': v, 'excludeFromIndexes': True}
    elif isinstance(v, float):
      p[k] = {'doubleValue': v, 'excludeFromIndexes': True}
    elif isinstance(v, bool):
      p[k] = {'booleanValue': v, 'excludeFromIndexes': True}
    elif isinstance(v, datetime.datetime):
      p[k] = {
          'timestampValue': v.strftime('%Y-%m-%dT%H:%M:%S.%fZ'),
          'excludeFromIndexes': True
      }
    else:
      raise Exception('No mapping from python to datastore for type of: %s' % k)

  return p


def datastore_write(auth, project_id, namespace, kind, key, values):
  """Writes a single row to datastore.

  For simplicity this is designed to store a single dictionary of values.

  Args
    - namespace (string): database
    - kind (string): table
    - key (string): primary key used for lookup
    - values (dictionary): record

  """

  body = {
      'mode':
          'NON_TRANSACTIONAL',
      'mutations': [{
          'upsert': {
              'key': {
                  'path': [{
                      'kind': kind,
                      'name': key
                  }],
                  'partitionId': {
                      'projectId': project_id,
                      'namespaceId': namespace
                  }
              },
              'properties': _datastore_v_to_p(values)
          }
      }]
  }

  API_Datastore(auth).projects().commit(
      projectId=project.id, body=body).execute()


def datastore_read(auth, project_id, namespace, kind, key):
  """Reads records from datastore based on supplied keys.

  Works on a single key, or an array of keys. Returns an iterator
  of tuples that can be cast to a list or a string:

  Sample usage:
    lookup_items = dict(datastore_read, "user", "some_project",
    "some_namespace", "some_kind", "some_key"))
    sort_items = list(datastore_read, "user", "some_project", "some_namespace",
    "some_kind", "some_key"))

  Args
    - namespace (string): database
    - kind (string): table
    - key (string or list): primary keys used for lookup

  Returns
    - key (string), values (dict): tuple iterator, even if single result
  """

  #response = API_Datastore(auth).projects().beginTransaction(
  #  projectId=project_id,
  #  body={
  #    "transactionOptions": { "readOnly":{} }
  #  }
  #).execute()

  # if single key givem turn it into a list
  if isinstance(key, str):
    key = [key]

  body = {
      'readOptions': {
          'readConsistency': 'STRONG'
      },
      'keys': [{
          'path': [{
              'kind': kind,
              'name': k
          }],
          'partitionId': {
              'projectId': project_id,
              'namespaceId': namespace
          }
      } for k in key]
  }

  response = API_Datastore(auth).projects().lookup(
      projectId=project.id, body=body).execute()

  # ignore missing, just do found for simplicity
  for e in response.get('found', []):
    yield _datastore_path(e['entity']['key']['path']), _datastore_p_to_v(
        e['entity']['properties'])


def datastore_list(auth, project_id, namespace, kind):
  """Reads all records from a datastore.

  Simplified query implementation that fetches all records. Designed for small
  non relational
  work.

  Sample usage:
    lookup_items = dict(datastore_list, "user", "some_project",
    "some_namespace", "some_kind"))
    sort_items = list(datastore_list, "user", "some_project", "some_namespace",
    "some_kind"))

  Args
    - namespace (string): database
    - kind (string): table

  Returns
    - key (string), values (dict): tuple iterator

  """

  body = {
      'partitionId': {
          'projectId': project_id,
          'namespaceId': namespace
      },
      'readOptions': {},
      'query': {
          'kind': [{
              'name': kind
          }]
      }
  }

  response = {'batch': {'moreResults': '', 'endCursor': None}}

  while response['batch']['moreResults'] != 'NO_MORE_RESULTS':
    body['query']['startCursor'] = response['batch']['endCursor']
    response = API_Datastore(auth).projects().runQuery(
        projectId=project.id, body=body).execute()

    for e in response['batch'].get('entityResults', []):
      yield _datastore_path(e['entity']['key']['path']), _datastore_p_to_v(
          e['entity']['properties'])



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

Conversion_Dimensions_Schema = [{
    'name': 'Conversion_Id',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Activity',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Activity_Group',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Activity_Group_Id',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Activity_Id',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Activity_Date_Time',
    'type': 'DATETIME',
    'mode': 'NULLABLE'
}, {
    'name': 'Browser_Platform',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Channel_Mix',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'City',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Click_Count',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Connection_Type',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Conversion_Referrer',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Conversion_Url',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Country',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Rich_Media_Custom_Event_Count',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Rich_Media_Custom_Event_Path_Summary',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Days_Since_Attributed_Interaction',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Days_Since_First_Interaction',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Designated_Market_Area_Dma',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Attribution_Type',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Configuration',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Hours_Since_Attributed_Interaction',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Hours_Since_First_Interaction',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Impression_Count',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Interaction_Count_Click_Tracker',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Interaction_Count_Mobile_Rich_Media',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Interaction_Count_Mobile_Static_Image',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Interaction_Count_Mobile_Video',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Interaction_Count_Natural_Search',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Interaction_Count_Paid_Search',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Interaction_Count_Rich_Media',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Interaction_Count_Static_Image',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Interaction_Count_Video',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Mobile_Carrier',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Num_Value',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Operating_System',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Operating_System_Version',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Ord_Value',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Path_Length',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Path_Type',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Platform_Type',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Recalculated_Attribution_Type',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Rich_Media_Standard_Event_Count',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Rich_Media_Standard_Event_Path_Summary',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'State_Region',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Tran_Value',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'U_Value',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Zip_Postal_Code',
    'type': 'STRING',
    'mode': 'NULLABLE'
}]



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

import re
import csv
import sys
from io import StringIO

from starthinker.util.project import project
from xlsx import Workbook

RE_HUMAN = re.compile('[^0-9a-zA-Z]+')
INT_LIMIT = 9223372036854775807  # defined by BigQuery 64 bit mostly ( not system )


def find_utf8_split(data, offset=None):

  if not isinstance(data, bytes):
    data = data.getbuffer()
  offset = offset or len(data)

  while offset > 0 and data[offset - 1] & 0xC0 == 0x80:
    offset -= 1

  if offset > 0:
    if data[offset - 1] & 0xE0 == 0xC0:
      offset -= 1
    elif data[offset - 1] & 0xF0 == 0xE0:
      offset -= 1
    elif data[offset - 1] & 0xF8 == 0xF0:
      offset -= 1

  return offset


def response_utf8_stream(response, chunksize):
  leftovers = b''

  while True:

    chunk = leftovers + response.read(chunksize)

    position = find_utf8_split(chunk)
    if position < len(chunk):
      leftovers = chunk[position:]
      chunk = chunk[:position]
    else:
      leftftovers = b''

    if chunk:
      yield chunk.decode('UTF-8')
    else:
      break


def bigquery_date(value):
  return value.strftime('%Y%m%d')


def excel_to_sheets(excel_file):
  excel_book = Workbook(excel_file)
  for excel_sheet in excel_book:
    yield excel_sheet.name


def excel_to_rows(excel_file, sheet=None):
  excel_book = Workbook(excel_file)
  # load all sheets in document
  for excel_sheet in excel_book:
    if sheet is None or sheet == excel_sheet.name:
      # load all rows from excel_sheet
      rows = []
      for row_number, cells in excel_sheet.rowsIter():
        rows.append(map(lambda cell: cell.value or cell.formula, cells))
      yield excel_sheet.name, rows


def csv_to_rows(csv_string):
  if csv_string:
    csv.field_size_limit(sys.maxsize)
    if isinstance(csv_string, str):
      csv_string = StringIO(csv_string)
    for row in csv.reader(
        csv_string,
        delimiter=',',
        quotechar='"',
        quoting=csv.QUOTE_MINIMAL,
        skipinitialspace=True,
        escapechar='\\'):
      yield row


def rows_to_csv(rows):
  csv_string = StringIO()
  writer = csv.writer(
      csv_string,
      delimiter=',',
      quotechar='"',
      quoting=csv.QUOTE_MINIMAL,
      lineterminator='\n')
  count = 0
  for row_number, row in enumerate(rows):
    try:
      writer.writerow(row)
      count += 1
    except Exception as e:
      print('Error:', row_number, str(e), row)
  csv_string.seek(0)  # important otherwise contents is zero
  if project.verbose:
    print('CSV Rows Written:', count)
  return csv_string


def rows_trim(rows):
  # find most common continous length
  histogram = {}
  prior = None
  for row in rows:
    length = len(row)
    if length != prior:
      histogram[length] = 1
    else:
      histogram[length] += 1
    prior = length
  common_length = sorted(
      histogram.iterkeys(), key=lambda k: histogram[k], reverse=True)[0]

  # strip any columns not in common length
  rows = [row for row in rows if len(row) == common_length]
  return rows


def rows_header_trim(rows):
  first = True
  for row in rows:
    if not first:
      yield row
    first = False


def column_header_sanitize(cell):
  header_sanitized = RE_HUMAN.sub('_',
                                  str(cell).title().replace(
                                      '%', 'Percent')).strip('_')
  if header_sanitized[0].isdigit():
    header_sanitized = '_' + header_sanitized  # bigquery does not take leading digits
  return header_sanitized


def row_header_sanitize(row):
  return [column_header_sanitize(cell) for cell in row]


def rows_header_sanitize(rows):
  first = True
  for row in rows:
    if first:
      row = row_header_sanitize(row)
      first = False
    yield row


def rows_percent_sanitize(rows):
  for row in rows:
    yield map(lambda c: c.replace('%', '').strip(), row)


def rows_date_sanitize(rows):
  first = True
  date = None
  for row in rows:
    if first:
      # find 'Date' column if it exists
      try:
        date = row.index('Date')
      except ValueError:
        pass

    # check if data studio formatting is applied
    if date is not None:
      row[date] = 'Report_Day' if first else row[date].replace('/', '').replace(
          '-', '')

    # return the row
    yield row

    # not first row anymore
    first = False


def rows_date_add(rows, date, header=True):
  if header:
    rows[0].insert(0, 'Report_Day')
    for row in rows[1:]:
      row.insert(0, bigquery_date(date))
  else:
    for row in rows:
      row.insert(0, bigquery_date(date))
  return rows


def rows_column_add(rows, header, value, index=None):
  first = True
  for row in rows:
    row.insert(
        index or len(row), header if first and header is not None else value)
    yield row
    first = False


def rows_column_delete(rows, column):
  for row in rows:
    del row[column:column + 1]
    yield row


def rows_null_to_value(rows, value):
  for row in rows:
    yield map(lambda c: value if c.strip().lower() == 'null' else c, row)


def rows_re_to_value(rows, regexp, value):
  regexp = re.compile(regexp)
  for row in rows:
    yield map(lambda c: regexp.sub(value, c), row)


def rows_to_type(rows, column=None):
  for row in rows:
    for index, value in enumerate(row):
      if column is None or column == index:
        # empty values are NULL ( avoid converting zero to null )
        if isinstance(value, str):
          if value == '':
            row[index] = None
          # all digits less than 64 bytes are integers
          elif value.isdigit():
            v = int(value)
            if abs(v) <= INT_LIMIT:
              row[index] = v
          # float probably needs a byte check
          elif '.' in value:
            w, d = value.split('.', 1)
            if w.isdigit() and d.isdigit():
              row[index] = float(value)
    yield row


def rows_print(rows, row_min=0, row_max=None):
  for i, row in enumerate(rows):
    if i >= row_min and (row_max is None or i <= row_max):
      print(i, row)
    yield row


def pivot_column_to_row(rows, discard_blanks=True):
  pivot = []

  for row in rows:
    for column, cell in enumerate(row):
      if len(pivot) == column:
        pivot.append([])
      if not discard_blanks or cell:
        pivot[column].append(cell)

  return pivot


def rows_to_dict(rows):
  return dict([(row[0], row[1:]) for row in rows])


def rows_slice(rows, row_min=0, row_max=100):
  for count, row in enumerate(rows):
    if count >= row_min:
      yield row
    if count >= row_max:
      break


def rows_pad(rows, length=0, padding=None):
  for row in rows:
    yield (row + [padding for i in range(0, length - len(row))]
          ) if len(row) < length else row



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

'''
--------------------------------------------------------------

Before running this Airflow module...

  Install StarThinker in cloud composer ( recommended ):

    From Release: pip install starthinker
    From Open Source: pip install git+https://github.com/google/starthinker

  Or push local code to the cloud composer plugins directory ( if pushing local code changes ):

    source install/deploy.sh
    4) Composer Menu
    l) Install All

--------------------------------------------------------------

  If any recipe task has "auth" set to "user" add user credentials:

    1. Ensure an RECIPE['setup']['auth']['user'] = [User Credentials JSON]

  OR

    1. Visit Airflow UI > Admin > Connections.
    2. Add an Entry called "starthinker_user", fill in the following fields. Last step paste JSON from authentication.
      - Conn Type: Google Cloud Platform
      - Project: Get from https://github.com/google/starthinker/blob/master/tutorials/cloud_project.md
      - Keyfile JSON: Get from: https://github.com/google/starthinker/blob/master/tutorials/deploy_commandline.md#optional-setup-user-credentials

--------------------------------------------------------------

  If any recipe task has "auth" set to "service" add service credentials:

    1. Ensure an RECIPE['setup']['auth']['service'] = [Service Credentials JSON]

  OR

    1. Visit Airflow UI > Admin > Connections.
    2. Add an Entry called "starthinker_service", fill in the following fields. Last step paste JSON from authentication.
      - Conn Type: Google Cloud Platform
      - Project: Get from https://github.com/google/starthinker/blob/master/tutorials/cloud_project.md
      - Keyfile JSON: Get from: https://github.com/google/starthinker/blob/master/tutorials/cloud_service.md

--------------------------------------------------------------

Census Data Normalized

Convert given census table to percent of population, normalizing it.

  - Specify the geography, year, and span that make up the <a href='https://pantheon.corp.google.com/bigquery?redirect_from_classic=true&p=bigquery-public-data&d=census_bureau_acs&page=dataset' target='_blank'>census table names</a>.
  - Not every combination of geography, year, and span exists or contains all the required fields.
  - Every time the query runs it will overwrite the table.

--------------------------------------------------------------

This StarThinker DAG can be extended with any additional tasks from the following sources:
  - https://google.github.io/starthinker/
  - https://github.com/google/starthinker/tree/master/dags

'''

from starthinker.airflow.factory import DAG_Factory

INPUTS = {
  'auth': 'service',  # Credentials used for writing data.
  'census_geography': 'zip_codes',  # Census table to get data from.
  'census_year': '2018',  # Census table to get data from.
  'census_span': '5yr',  # Census table to get data from.
  'dataset': '',  # Existing BigQuery dataset.
  'type': 'table',  # Write Census_Percent as table or view.
}

RECIPE = {
  'tasks': [
    {
      'census': {
        'auth': {
          'field': {
            'name': 'auth',
            'kind': 'authentication',
            'order': 0,
            'default': 'service',
            'description': 'Credentials used for writing data.'
          }
        },
        'normalize': {
          'census_geography': {
            'field': {
              'name': 'census_geography',
              'kind': 'choice',
              'order': 1,
              'default': 'zip_codes',
              'description': 'Census table to get data from.',
              'choices': [
                'zip_codes',
                'state',
                'zcta5',
                'schooldistrictunified',
                'puma',
                'place',
                'county',
                'congressionaldistrict',
                'censustract',
                'cbsa'
              ]
            }
          },
          'census_year': {
            'field': {
              'name': 'census_year',
              'kind': 'choice',
              'order': 2,
              'default': '2018',
              'description': 'Census table to get data from.',
              'choices': [
                2018,
                2017,
                2016,
                2015,
                2014,
                2013,
                2012,
                2011,
                2010,
                2009,
                2008,
                2007
              ]
            }
          },
          'census_span': {
            'field': {
              'name': 'census_span',
              'kind': 'choice',
              'order': 3,
              'default': '5yr',
              'description': 'Census table to get data from.',
              'choices': [
                '1yr',
                '3yr',
                '5yr'
              ]
            }
          }
        },
        'to': {
          'dataset': {
            'field': {
              'name': 'dataset',
              'kind': 'string',
              'order': 4,
              'default': '',
              'description': 'Existing BigQuery dataset.'
            }
          },
          'type': {
            'field': {
              'name': 'type',
              'kind': 'choice',
              'order': 5,
              'default': 'table',
              'description': 'Write Census_Percent as table or view.',
              'choices': [
                'table',
                'view'
              ]
            }
          }
        }
      }
    }
  ]
}

DAG_FACTORY = DAG_Factory('bigquery_census_normalize', RECIPE, INPUTS)
DAG = DAG_FACTORY.generate()

if __name__ == "__main__":
  DAG_FACTORY.print_commandline()



import json
import os
import csv
import uuid

from starthinker.util.sheets import sheets_read, sheets_id
from starthinker.util.project import project
from starthinker.util.bigquery import io_to_table, get_schema, query_to_rows

output_file_name = '/tmp/%s' % str(uuid.uuid1())


def _load_admin_table():
  if project.verbose:
    print('Omnicom Admin Trix')

  # Read Admin Sheet (format: Name, Sheet URL, Partner ID, Advertiser IDs)
  client_sheet_ids = []
  split = project.task['range'].split('!')
  sheet_name = split[0]
  sheet_range = split[1]

  admin_sheet = sheets_read(project.task['auth'], project.task['sheet_id'],
                            sheet_name, sheet_range)

  output_file = open(output_file_name, 'wb')
  output = csv.writer(output_file)

  # Admin table header
  output.writerow(['Name', 'Sheet_ID', 'Partner_ID', 'Advertiser_ID'])

  # For each row
  for admin_row in admin_sheet[1:]:
    # Parse Sheet ID
    client_sheet_id = sheets_id(project.task['auth'], admin_row[1])
    client_sheet_ids.append(client_sheet_id)

    base_row = [admin_row[0], client_sheet_id, admin_row[2]]

    # If advertiser ids are not present, it means all advertisers under the
    # partner, so pull advertisers from BQ
    if len(admin_row) >= 4 and admin_row[3]:
      advertisers = [
          int(advertiser.strip()) for advertiser in admin_row[3].split(',')
      ]
    else:
      advertisers = [
          advertiser[0] for advertiser in query_to_rows(
              project.task['auth'],
              project.id,
              project.task['dataset'],
              'select advertiser_id from starthinker_v2.Advertiser where partner_id = %d'
              % int(admin_row[2]),
              row_max=None,
              legacy=True)
      ]

    for advertiser_id in advertisers:
      # For each advertiser, add row to output (format: Name, Sheet ID, Partner ID, Advertiser ID)
      output.writerow(base_row + [advertiser_id])

  output_file.close()

  # Load Admin table, overwriting existing content
  rows, schema = get_schema(
      [['Name', 'Sheet_ID', 'Partner_ID', 'Advertiser_ID']], True, True)
  io_to_table(
      project.task['auth'],
      project.id,
      project.task['dataset'],
      project.task['table'],
      open(output_file_name, 'rb'),
      'CSV',
      schema=schema,
      skip_rows=1)

  os.remove(output_file_name)

  return client_sheet_ids


def _load_client_sheet(client_sheet_ids, sheet_name, sheet_range, table_name):
  output_file = open(output_file_name, 'wb')
  output = csv.writer(output_file)

  # For each input sheet
  for client_sheet_id in client_sheet_ids:

    # Reed the input sheet
    client_sheet = sheets_read(project.task['auth'], client_sheet_id,
                               sheet_name, sheet_range)

    # Add header with additional Sheet ID column
    output.writerow(client_sheet[0] + ['Sheet_ID'])

    # Inject sheet id in each row, and add to output
    for row in client_sheet[1:]:
      output.writerow(row + [client_sheet_id])

  # Upload to Big Query
  output_file.close()

  rows, schema = get_schema([client_sheet[0] + ['Sheet_ID']], True, True)
  io_to_table(
      project.task['auth'],
      project.id,
      project.task['dataset'],
      table_name,
      open(output_file_name, 'rb'),
      'CSV',
      schema=schema,
      skip_rows=1)

  #os.remove(output_file_name)


@project.from_parameters
def omnicom_admin_trix():
  # Step1: Load admin trix replacing the sheet URL with just the sheet ID, and
  # splitting the advertisers
  client_sheet_ids = _load_admin_table()

  # Step 2: Load each client specific sheet injecting the sheet ID in each row
  _load_client_sheet(client_sheet_ids, 'Traders', 'A1:C', 'traders')
  _load_client_sheet(client_sheet_ids, 'Impression_Goals', 'A1:C',
                     'impression_goals')
  _load_client_sheet(client_sheet_ids, 'Advertisers', 'A1:E',
                     'advertiser_budget')


if __name__ == '__main__':
  omnicom_admin_trix()



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

import re
import json
import base64
from time import sleep
from io import BytesIO

from googleapiclient import discovery
from googleapiclient.http import MediaIoBaseUpload
from googleapiclient.errors import HttpError
from google.oauth2.service_account import Credentials

from starthinker.config import UI_SERVICE

RE_CREDENTIALS_JSON = re.compile(r'^\s*\{.*\}\s*$', re.DOTALL)


def _credentials_storage_service():

  if RE_CREDENTIALS_JSON.match(UI_SERVICE):
    credentials = Credentials.from_service_account_info(json.loads(UI_SERVICE))
  else:
    credentials = Credentials.from_service_account_file(UI_SERVICE)

  return discovery.build('storage', 'v1', credentials=credentials)


def _credentials_retry(job, retries=3, wait=1):
  try:
    return job.execute()
  except HttpError as e:
    if e.resp.status == 429 and retries > 0:
      sleep(wait)
      return _credentials_retry(job, retries - 1, wait * 2)
    else:
      raise


def credentials_storage_get(cloud_path):
  bucket, filename = cloud_path.split(':', 1)
  data = _credentials_retry(_credentials_storage_service().objects().get_media(
      bucket=bucket, object=filename))
  return json.loads(base64.b64decode(data.decode()).decode())


def credentials_storage_put(cloud_path, credentials):
  bucket, filename = cloud_path.split(':', 1)
  data = BytesIO(base64.b64encode(json.dumps(credentials).encode()))
  media = MediaIoBaseUpload(data, mimetype='text/json')
  _credentials_retry(_credentials_storage_service().objects().insert(
      bucket=bucket, name=filename, media_body=media))



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

from starthinker.util.project import project
from starthinker.util.data import put_rows, get_rows
from starthinker.util.dbm import lineitem_get_v1, lineitem_patch_v1
from starthinker.util.dbm.schema import LineItem_Read_Schema


def handle_entity_status(input_li, dv_li):
  """Handles DV360 entity status field by applying changes from the input_li to the dv_li

  Args:
    input_li: object representing the input line item from the input data
    dv_li: object representing the DV360 line item
  Returns: List of strings representing the fields to patch
  """
  if 'Active' in input_li:
    if input_li['Active'] != '':
      dv_li['entityStatus'] = 'ENTITY_STATUS_ACTIVE' if input_li[
          'Active'] == 'TRUE' else 'ENTITY_STATUS_PAUSED'
      return ['entityStatus']

  return []


def handle_fixed_bid(input_li, dv_li):
  """Handles DV360 fixed bid field by applying changes from the input_li to the dv_li

  Args:
    input_li: object representing the input line item from the input data
    dv_li: object representing the DV360 line item
  Returns: List of strings representing the fields to patch
  """
  if 'Fixed Bid' in input_li:
    if input_li['Fixed Bid'] != '':
      dv_li['bidStrategy'] = {
          'fixedBid': {
              'bidAmountMicros': float(input_li['Fixed Bid']) * 1000000
          }
      }
      return ['bidStrategy.fixedBid.bidAmountMicros']

  return []


def handle_patch(input_li, dv_li):
  """Handles that patch operation by leveraging the handler functions to apply changes represented by the input_li to the dv_li

  Args:
    input_li: object representing the input line item from the input data
    dv_li: object representing the DV360 line item
  Returns: List of strings representing the fields to patch
  """
  result = []

  result += handle_entity_status(input_li, dv_li)
  result += handle_fixed_bid(input_li, dv_li)

  return result


@project.from_parameters
def lineitem_beta():
  """Main entry point of the lineitem_beta StarThinker task, here is the task definition:{

      "lineitem_beta":{
        "auth":"user",
        "read":{
          "sheet":{
            "sheet":
            "https://docs.google.com/spreadsheets/d/109jJpQa6QUIjtoY6wI5yPAkixzrmh33q9AE-zjwcsTE/edit#gid=0",
            "tab": "Rules",
            "range": "A1:D"
          }
        },
        "patch": {}
      }
    }

  read: represents the input data for operating in the line items
  patch: indicates a patch operation should be performed
  """
  if project.verbose:
    print('LINEITEM_BETA')

  li_map = {}

  if 'read' in project.task:
    #TODO this has been developed to work with sheets in this initial phase,
    # need to check how to handle headers in case of reading from other sources such as BQ
    rows = get_rows(project.task['auth'], project.task['read'])

    header = False

    for row in rows:
      if not header:
        header = row
      else:
        li = {}
        for idx, field in enumerate(header):
          li[field] = row[idx]

        dv_li = lineitem_get_v1(project.task['auth'], li['Advertiser ID'],
                                li['Line Item ID'])

        li_map[li['Line Item ID']] = {'input_li': li, 'dv_li': dv_li}

  if 'patch' in project.task:
    patch = []

    for li_id in li_map.keys():
      input_li = li_map[li_id]['input_li']
      dv_li = li_map[li_id]['dv_li']

      patch = handle_patch(input_li, dv_li)

      if patch:
        lineitem_patch_v1(project.task['auth'], ','.join(patch), dv_li)


if __name__ == '__main__':
  lineitem_beta()



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

from datetime import datetime, date, timedelta

from starthinker.util.project import project
from starthinker.util.storage import object_list, object_move, object_delete


@project.from_parameters
def archive():
  if project.verbose:
    print('ARCHIVE')

  day = project.date - timedelta(days=abs(project.task['days']))

  for object in object_list(
      project.task['auth'],
      project.task['storage']['bucket'] + ':' + project.task['storage']['path'],
      files_only=True,
      raw=True):
    object_day = datetime.strptime(object['updated'],
                                   '%Y-%m-%dT%H:%M:%S.%fZ').date()
    if object_day <= day:
      if project.task.get('delete', False) == False:
        if project.verbose:
          print('ARCHIVING FILE:', object['name'])
        object_move(project.task['auth'],
                    '%s:%s' % (object['bucket'], object['name']),
                    '%s:archive/%s' % (object['bucket'], object['name']))
      else:
        if project.verbose:
          print('DELETING FILE:',)
        object_delete(project.task['auth'],
                      '%s:%s' % (object['bucket'], object['name']))


if __name__ == '__main__':
  archive()



#!/usr/bin/env python

###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

import os
import sys

if __name__ == '__main__':
  os.environ.setdefault('DJANGO_SETTINGS_MODULE', 'ui.settings')
  try:
    from django.core.management import execute_from_command_line
  except ImportError:
    # The above import may fail for some other reason. Ensure that the
    # issue is really that Django is missing to avoid masking other
    # exceptions on Python 2.
    try:
      import django
    except ImportError:
      raise ImportError(
          "Couldn't import Django. Are you sure it's installed and "
          'available on your PYTHONPATH environment variable? Did you '
          'forget to activate a virtual environment?')
    raise
  execute_from_command_line(sys.argv)



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

import os
import errno
import json
import httplib2
from time import sleep
from io import BytesIO

from googleapiclient.http import MediaIoBaseUpload, MediaIoBaseDownload
from googleapiclient.errors import HttpError

from starthinker.config import BUFFER_SCALE
from starthinker.util.project import project
from starthinker.util.google_api import API_Storage
from starthinker.util.csv import find_utf8_split

CHUNKSIZE = int(200 * 1024000 *
                BUFFER_SCALE)  # scale is controlled in config.py
RETRIES = 3


def makedirs_safe(path):
  try:
    os.makedirs(path)
  except OSError as exc:
    if exc.errno == errno.EEXIST and os.path.isdir(path):
      pass
    else:
      raise


def parse_path(path):
  try:
    return path.rsplit('/', 1)[0]
  except:
    return ''


def parse_filename(path, url=False):
  f = path
  try:
    if url:
      f = f.split('?', 1)[0]
    f = f.rsplit('/', 1)[1]
  except:
    pass
  return f


def media_download(request, chunksize, encoding=None):
  data = BytesIO()
  leftovers = b''

  media = MediaIoBaseDownload(data, request, chunksize=chunksize)

  retries = 0
  done = False
  while not done:
    error = None
    try:
      progress, done = media.next_chunk()
      if progress:
        print('Download %d%%' % int(progress.progress() * 100))

      data.seek(0)

      if encoding is None:
        yield data.read()

      elif encoding.lower() == 'utf-8':
        position = find_utf8_split(data)
        yield (leftovers + data.read(position)).decode(encoding)
        leftftovers = data.read()

      else:
        yield data.read().decode(encoding)

      data.seek(0)
      data.truncate(0)
    except HttpError as err:
      error = err
      if err.resp.status < 500:
        raise
    except (httplib2.HttpLib2Error, IOError) as err:
      error = err

    if error:
      retries += 1
      if retries > RETRIES:
        raise error
      else:
        sleep(5 * retries)
    else:
      retries = 0

  print('Download 100%')


def object_exists(auth, path):
  bucket, filename = path.split(':', 1)
  try:
    API_Storage(auth).objects().get(bucket=bucket, object=filename).execute()
    return True
  except:
    return False


def object_get(auth, path):
  bucket, filename = path.split(':', 1)
  return API_Storage(auth).objects().get_media(bucket=bucket, object=filename).execute()


def object_get_chunks(auth, path, chunksize=CHUNKSIZE, encoding=None):
  bucket, filename = path.split(':', 1)
  data = BytesIO()
  request = API_Storage(auth).objects().get_media(bucket=bucket, object=filename).execute(run=False)
  yield from media_download(request, chunksize, encoding)


def object_put(auth, path, data, mimetype='application/octet-stream'):
  bucket, filename = path.split(':', 1)

  media = MediaIoBaseUpload(
      data, mimetype=mimetype, chunksize=CHUNKSIZE, resumable=True)
  request = API_Storage(auth).objects().insert(
      bucket=bucket, name=filename, media_body=media).execute(run=False)

  response = None
  errors = 0
  while response is None:
    error = None
    try:
      status, response = request.next_chunk()
      if project.verbose and status:
        print('Uploaded %d%%.' % int(status.progress() * 100))
    except HttpError as e:
      if e.resp.status < 500:
        raise
      error = e
    except (httplib2.HttpLib2Error, IOError) as e:
      error = e

    errors = (errors + 1) if error else 0
    if errors > RETRIES:
      raise error

  if project.verbose:
    print('Uploaded 100%.')


def object_list(auth, path, raw=False, files_only=False):
  bucket, prefix = path.split(':', 1)
  for item in API_Storage(
      auth, iterate=True).objects().list(
          bucket=bucket, prefix=prefix).execute():
    if files_only and item['name'].endswith('/'):
      continue
    yield item if raw else '%s:%s' % (bucket, item['name'])


def object_copy(auth, path_from, path_to):
  from_bucket, from_filename = path_from.split(':', 1)
  to_bucket, to_filename = path_to.split(':', 1)

  body = {
      'kind': 'storage#object',
      'bucket': to_bucket,
      'name': to_filename,
      'storageClass': 'REGIONAL',
  }

  return API_Storage(auth).objects().rewrite(
      sourceBucket=from_bucket,
      sourceObject=from_filename,
      destinationBucket=to_bucket,
      destinationObject=to_filename,
      body=body
  ).execute()


def object_delete(auth, path):
  bucket, filename = path.split(':', 1)
  return API_Storage(auth).objects().delete(bucket=bucket, object=filename).execute()


def object_move(auth, path_from, path_to):
  object_copy(auth, path_from, path_to)
  object_delete(auth, path_from)


def bucket_get(auth, name):
  try:
    return API_Storage(auth).buckets().get(bucket=name).execute()
  except HttpError as e:
    if e.resp.status == 404:
      return None
    elif e.resp.status in [403, 500, 503]:
      sleep(5)
    else:
      raise


def bucket_create(auth, project, name, location='us-west1'):
  if bucket_get(auth, name) is None:
    body = {
        'kind': 'storage#bucket',
        'name': name,
        'storageClass': 'REGIONAL',
        'location': location,
    }

    try:
      return API_Storage(auth).buckets().insert(project=project, body=body).execute()
      sleep(1)
    except HttpError as e:
      if e.resp.status in [403, 500, 503]:
        sleep(5)
      elif json.loads(e.content.decode())['error']['code'] == 409:
        pass  # already exists ( ignore )
      else:
        raise


def bucket_delete(auth, name):
  return API_Storage(auth).buckets().delete(bucket=name).execute()


#role = OWNER, READER, WRITER
def bucket_access(auth,
                  project,
                  name,
                  role,
                  emails=[],
                  groups=[],
                  services=[],
                  domains=[]):

  entities = map(lambda e: 'user-%s' % e, emails) + \
    map(lambda e: 'group-%s' % e, groups) + \
    map(lambda e: 'user-%s' % e, services) + \
    map(lambda e: 'domain-%s' % e, domains)

  for entity in entities:
    body = {
        'kind': 'storage#bucketAccessControl',
        'bucket': name,
        'entity': entity,
        'role': role
    }
    API_Storage(auth).bucketAccessControls().insert(bucket=name, body=body).execute()


# Alternative for managing permissions ( overkill? )
#  if emails or groups or services or groups:
#    access = service.buckets().getIamPolicy(bucket=name).execute(num_retries=RETRIES)
#
#    access['bindings'] = []
#    for r in role:
#      access['bindings'].append({
#        "role":"roles/storage.object%s" % r,
#        "members": ['user:%s' % m for m in emails] + \
#                   ['group:%s' % m for m in groups] + \
#                   ['serviceAccount:%s' % m for m in services] + \
#                   ['domain:%s' % m for m in domains]
#      })
#
#    job = service.buckets().setIamPolicy(bucket=name, body=access).execute(num_retries=RETRIES)
#    sleep(1)



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

import os
import json
import shutil
from time import sleep
from datetime import datetime, timedelta
from django.test import TestCase

from starthinker.util.project import project
from starthinker.util.auth import clear_credentials_cache, get_credentials, get_service
from starthinker.config import UI_CLIENT, UI_SERVICE
from starthinker_ui.account.models import Account

UI_USER = os.environ.get('STARTHINKER_USER', '')


def account_create():

  accounts = Account.objects.all()
  if len(accounts) > 0:
    account = accounts[0]
  else:
    project.initialize(_client=UI_CLIENT, _service=UI_SERVICE, _user=UI_USER)
    credentials = get_credentials('user')
    account = Account.objects.get_or_create_user(credentials, 'password')

  return account


class CredentialsTest(TestCase):

  def setUp(self):
    self.user_file = '/tmp/test_user.json'
    self.service_file = '/tmp/test_service.json'
    shutil.copyfile(UI_USER, self.user_file)
    shutil.copyfile(UI_SERVICE, self.service_file)

  def tearDown(self):
    os.remove(self.user_file)
    os.remove(self.service_file)

  def helper_refresh(self):
    credentials = get_credentials('user')
    token = credentials.token
    expiry = credentials.expiry

    # wait a bit before refreshing token, multiple tests go too fast and same token is returned
    sleep(1)

    # test refresh ( not expired cache, not expired file )
    credentials.refresh()
    self.assertEqual(token, credentials.token)
    self.assertEqual(expiry, credentials.expiry)

    # wait a bit before refreshing token, multiple tests go too fast and same token is returned
    sleep(1)

    # test refresh ( expired cache, not expired file )
    credentials.expiry = (datetime.now() - timedelta(days=5))
    credentials.refresh()
    self.assertEqual(token, credentials.token)
    self.assertEqual(expiry, credentials.expiry)

    # wait a bit before refreshing token, multiple tests go too fast and same token is returned
    sleep(1)

    # test refresh ( expired cache, expired file )
    credentials.expiry = (datetime.now() - timedelta(days=5))
    credentials.save()
    credentials.refresh()
    self.assertNotEqual(token, credentials.token)
    self.assertNotEqual(expiry, credentials.expiry)

  def test_file_credentials_user(self):
    project.initialize(_user=self.user_file)

    service = get_service('oauth2', 'v2', 'user')
    response = service.userinfo().get().execute()

    self.assertIn('email', response)
    self.helper_refresh()

  def test_file_credentials_service(self):
    project.initialize(_service=self.service_file)

    service = get_service('cloudresourcemanager', 'v1', 'service')
    response = service.projects().list().execute()

    self.assertIn('projects', response)

  def test_string_credentials_user(self):
    with open(self.user_file, 'r') as json_file:
      project.initialize(_user=json_file.read())

    service = get_service('oauth2', 'v2', 'user')
    response = service.userinfo().get().execute()

    self.assertIn('email', response)
    self.helper_refresh()

  def test_string_credentials_service(self):
    with open(self.service_file, 'r') as json_file:
      project.initialize(_service=json_file.read())

    service = get_service('cloudresourcemanager', 'v1', 'service')
    response = service.projects().list().execute()

    self.assertIn('projects', response)

  def test_dictionary_credentials_user(self):
    with open(self.user_file, 'r') as json_file:
      project.initialize(_user=json.load(json_file))

    service = get_service('oauth2', 'v2', 'user')
    response = service.userinfo().get().execute()

    self.assertIn('email', response)
    self.helper_refresh()

  def test_dictionary_credentials_service(self):
    with open(self.service_file, 'r') as json_file:
      project.initialize(_service=json.load(json_file))

    service = get_service('cloudresourcemanager', 'v1', 'service')
    response = service.projects().list().execute()

    self.assertIn('projects', response)

  def test_remote_credentials_user(self):
    project.initialize(_user=self.user_file)
    credentials = get_credentials('user')
    account = Account.objects.get_or_create_user(credentials, 'password')

    clear_credentials_cache()

    project.initialize(_user=account.get_credentials_path())
    self.assertEqual(project.recipe['setup']['auth']['user'],
                     account.get_credentials_path())

    service = get_service('oauth2', 'v2', 'user')
    response = service.userinfo().get().execute()

    self.assertIn('email', response)
    self.helper_refresh()



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

Paid_Search_Schema = [{
    'name': 'Ad_Id',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Advertiser_Id',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Campaign_Id',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Paid_Search_Ad_Group',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Paid_Search_Ad_Group_Id',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Paid_Search_Ad_Id',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Paid_Search_Bid_Strategy',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Paid_Search_Bid_Strategy_Id',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Paid_Search_Campaign',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Paid_Search_Campaign_Id',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Paid_Search_Landing_Page_Url',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Paid_Search_Legacy_Keyword_Id',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Paid_Search_Keyword',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Paid_Search_Keyword_Id',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Paid_Search_Match_Type',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Config_Id',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}]



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

from starthinker.util.project import project
from starthinker.util.bigquery import query_to_table, query_to_view
from starthinker.task.census.fields import CENSUS_FIELDS


def census_normalize(census_geography, census_year, census_span):

  query = 'SELECT\n'
  for segment in CENSUS_FIELDS:
    if segment['denominator'] is None:
      for column in segment['columns']:
        query += '  %s AS %s,\n' % (column, column.title())
    else:
      query += '\n  /* %s */\n' % segment['category']
      for column in segment['columns']:
        query += '  SAFE_DIVIDE(%s, %s) AS %s,\n' % (
            column, segment['denominator'], column.title())

  query += 'FROM `bigquery-public-data.census_bureau_acs.%s_%s_%s`;' % (
      census_geography, census_year, census_span)

  return query


def census_pivot(dataset, table):
  query = 'SELECT Geo_Id, [\n'

  for segment in CENSUS_FIELDS:
    if segment['category']:
      for column_a in segment['columns']:
        query += '    STRUCT(\n'
        query += '      "%s" AS Percentage,\n' % column_a.title()
        query += '      "%s" AS Dimension,\n' % segment['category']
        query += '      "%s" AS Segment\n' % column_a.title()
        query += '      ),\n'

  query = query[:-2] + '\n  ] AS Correlation\n'
  query += 'FROM `%s.%s`;' % (dataset, table)

  return query


def census_correlate(column_join, column_pass, column_sum, column_correlate,
                     dataset, table):

  query = 'SELECT\n'

  for column in column_pass:
    query += '  A.%s,\n' % column

  for column in column_sum:
    query += '  SUM(A.%s) AS %s,\n' % (column, column)

  query += '  COUNT(*) AS Locations,\n'

  for segment in CENSUS_FIELDS:
    if segment['category']:
      for column_a in segment['columns']:
        for column_b in column_correlate:
          query += '  CORR(A.%s, B.%s) AS %s_%s_Correlation,\n' % (
              column_b, column_a.title(), column_a.title(), column_b)

  query += 'FROM `%s.%s.%s` AS A\n' % (project.id, dataset, table)
  query += 'LEFT JOIN `%s.%s.Census_Normalized` AS B\n' % (project.id, dataset)
  query += 'ON A.%s=B.Geo_Id\n' % column_join
  query += 'GROUP BY %s' % ','.join(map(str, range(1, len(column_pass) + 1)))

  return query


def census_significance(column_join, column_pass, column_sum, column_correlate,
                        dataset, significance):

  query = 'SELECT\n'

  for column in column_pass:
    query += '  %s,\n' % column

  for column in column_sum:
    query += '  %s,\n' % column

  query += '  Locations,\n'

  for segment in CENSUS_FIELDS:
    if segment['category']:
      for column_a in segment['columns']:
        for column_b in column_correlate:
          query += ('%s.pearson_significance_test(%s_%s_Correlation, Locations,'
                    ' %s, 0.5) AS %s_%s_Correlation,\n') % (
              dataset,
              column_a.title(),
              column_b,
              significance,
              column_a.title(),
              column_b,
          )

  query += 'FROM `%s.%s.Census_Correlated`;' % (project.id, dataset)

  return query


def census_join(column_join, column_pass, column_sum, column_correlate,
                 dataset):

  query = 'SELECT\n'

  for column in column_pass:
    query += '  %s,\n' % column

  for column in column_sum:
    query += '  %s,\n' % column

  query += '  Locations,\n'

  query += '  [\n'

  for segment in CENSUS_FIELDS:
    if segment['category']:
      for column_a in segment['columns']:
        query += '    STRUCT(\n'
        for column_b in column_correlate:
          query += '      %s_%s_Correlation AS %s,\n' % (column_a.title(),
                                                         column_b, column_b)
        query += '      "%s" AS Dimension,\n' % segment['category']
        query += '      "%s" AS Segment\n' % column_a.title()
        query += '      ),\n'

  query = query[:-2] + '\n  ] AS Correlation\n'
  query += 'FROM `%s.%s.Census_Significant`;' % (project.id, dataset)

  return query


def census_write(query, table):

  if project.verbose:
    print('%s: %s' % (table, query))

  if project.task['to']['type'] == 'table':
    query_to_table(
        project.task['auth'],
        project.id,
        project.task['to']['dataset'],
        table,
        query,
        legacy=False)

  elif project.task['to']['type'] == 'view':
    query_to_view(
        project.task['auth'],
        project.id,
        project.task['to']['dataset'],
        table,
        query,
        legacy=False)


@project.from_parameters
def census():
  if 'normalize' in project.task:
    query = census_normalize(
        project.task['normalize']['census_geography'],
        project.task['normalize']['census_year'],
        project.task['normalize']['census_span'],
    )
    census_write(query, 'Census_Normalized')

  if project.task.get('pivot'):
    query = census_pivot(project.task['to']['dataset'], 'Census_Normalized')
    census_write(query, 'Census_Pivoted')

  if 'correlate' in project.task:
    query = census_correlate(
        project.task['correlate']['join'],
        project.task['correlate']['pass'],
        project.task['correlate']['sum'],
        project.task['correlate']['correlate'],
        project.task['correlate']['dataset'],
        project.task['correlate']['table'],
    )
    census_write(query, 'Census_Correlated')

    query = census_significance(
        project.task['correlate']['join'],
        project.task['correlate']['pass'],
        project.task['correlate']['sum'],
        project.task['correlate']['correlate'],
        project.task['correlate']['dataset'],
        project.task['correlate']['significance'],
    )
    census_write(query, 'Census_Significant')

    query = census_join(project.task['correlate']['join'],
                         project.task['correlate']['pass'],
                         project.task['correlate']['sum'],
                         project.task['correlate']['correlate'],
                         project.task['correlate']['dataset'])
    census_write(query, 'Census_Join')


#WITH data AS (
#  SELECT ST_GEOGPOINT(-122.4194, 37.7749) point
#  UNION ALL
#  SELECT ST_GEOGPOINT(-74.0060, 40.7128) point
#)
#SELECT zip_code, city, county
#FROM `bigquery-public-data.geo_us_boundaries.zip_codes`
#JOIN data
#ON ST_WITHIN(point, zip_code_geom)

if __name__ == '__main__':
  census()



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

#https://cloud.google.com/compute/docs/reference/rest/v1/instanceGroupManagers

from django.conf import settings
from starthinker.util.project import project
from starthinker.util.google_api import API_Compute


def group_instances_resize(count):
  project.initialize(_service=settings.UI_SERVICE, _project=settings.UI_PROJECT)
  return API_Compute('service').instanceGroupManagers().resize(
      project=settings.UI_PROJECT,
      zone=settings.UI_ZONE,
      instanceGroupManager='starthinker-worker-group',
      size=count).execute()


def group_instances_delete(name):
  project.initialize(_service=settings.UI_SERVICE, _project=settings.UI_PROJECT)
  return API_Compute('service').instanceGroupManagers().deleteInstances(
      project=settings.UI_PROJECT,
      zone=settings.UI_ZONE,
      instanceGroupManager='starthinker-worker-group',
      body={
          'instances': ['zones/%s/instances/%s' % (settings.UI_ZONE, name)],
          'type': 'PROACTIVE'
      }).execute()


def group_instances_list(statuses=[]):
  project.initialize(_service=settings.UI_SERVICE, _project=settings.UI_PROJECT)
  return API_Compute(
      'service', iterate=True).instanceGroupManagers().listManagedInstances(
          project=settings.UI_PROJECT,
          zone=settings.UI_ZONE,
          instanceGroupManager='starthinker-worker-group',
          filter=' OR '.join(
              ['(instanceStatus = "%s")' % status for status in statuses]),
          orderBy='creationTimestamp desc').execute()



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

DT_Field_Lookup = {
    'Zip_Postal_Code': 'STRING',
    'Tag_Counting_Method_Id': 'INTEGER',
    'Event_Sub_Type': 'STRING',
    'Account_Id': 'INTEGER',
    'Creative_Type_Id': 'INTEGER',
    'Dbm_Total_Media_Cost_Advertiser_Currency': 'INTEGER',
    'Content_Category': 'STRING',
    'Dbm_Cpm_Fee_4_Usd': 'FLOAT',
    'Country_Code': 'STRING',
    'Rich_Media_Event_Type_Id': 'INTEGER',
    'Partitiontime': 'TIMESTAMP',
    'Partner_2_Id': 'STRING',
    'Dbm_Cpm_Fee_5_Advertiser_Currency': 'FLOAT',
    'Dbm_Media_Fee_4_Usd': 'FLOAT',
    'Billing_Invoice_Code': 'STRING',
    'Package_Roadblock_Type_Id': 'INTEGER',
    'Dbm_Browser_Platform_Id': 'INTEGER',
    'Dbm_Media_Fee_3_Partner_Currency': 'FLOAT',
    'Dbm_Revenue_Advertiser_Currency': 'FLOAT',
    'Site_Keyname': 'STRING',
    'Site_Id_Dcm': 'INTEGER',
    'Custom_Event_Timers': 'INTEGER',
    'Event_Timers': 'FLOAT',
    'Dbm_Country_Code': 'STRING',
    'Dbm_Media_Fee_5_Usd': 'FLOAT',
    'Dbm_Media_Fee_5_Partner_Currency': 'FLOAT',
    'Activity_Group': 'STRING',
    'Creative_Field_Number': 'INTEGER',
    'Floodlight_Variable_Name': 'STRING',
    'Paid_Search_Landing_Page_Url': 'STRING',
    'Dbm_Media_Fee_3_Advertiser_Currency': 'FLOAT',
    'Keyword': 'STRING',
    'Dbm_Request_Time': 'INTEGER',
    'Region': 'STRING',
    'Dma_Region': 'STRING',
    'Site_Id': 'INTEGER',
    'Placement_Cap_Cost_Option': 'STRING',
    'Dbm_Media_Fee_3_Usd': 'FLOAT',
    'Advertiser_Id': 'INTEGER',
    'Dbm_Media_Cost_Usd': 'INTEGER',
    'Creative_Version': 'STRING',
    'Creative_Size': 'STRING',
    'Creative': 'STRING',
    'Package_Roadblock_Type': 'STRING',
    'Paid_Search_Campaign': 'STRING',
    'Dbm_Total_Media_Cost_Partner_Currency': 'INTEGER',
    'Campaign': 'STRING',
    'Event_Time': 'INTEGER',
    'Dbm_Cpm_Fee_4_Advertiser_Currency': 'FLOAT',
    'Creative_Type': 'STRING',
    'Rendering_Id': 'INTEGER',
    'Dbm_Media_Fee_4_Partner_Currency': 'FLOAT',
    'Package_Roadblock_Id': 'INTEGER',
    'Content_Category_Id': 'INTEGER',
    'Referrer_Url': 'STRING',
    'Other_Data': 'STRING',
    'Dbm_Advertiser_Id': 'INTEGER',
    'Creative_Group2_Id': 'INTEGER',
    'Total_Conversions': 'INTEGER',
    'Paid_Search_Keyword': 'STRING',
    'Floodlight_Configuration': 'INTEGER',
    'Ad_Id': 'INTEGER',
    'Interaction_Time': 'STRING',
    'Flighting_Activated': 'BOOLEAN',
    'Event_Counters': 'INTEGER',
    'Region_Full_Name': 'STRING',
    'Landing_Page_Url': 'STRING',
    'Dbm_Mobile_Make_Id': 'INTEGER',
    'Active_View_Eligible_Impressions': 'INTEGER',
    'Browser_Platform_Version': 'STRING',
    'Creative_Start_Date': 'INTEGER',
    'Dbm_Cpm_Fee_3_Usd': 'FLOAT',
    'Dbm_Cpm_Fee_2_Advertiser_Currency': 'FLOAT',
    'Dbm_Revenue_Usd': 'INTEGER',
    'Advertiser_Group_Id': 'INTEGER',
    'Creative_Group1_Id': 'INTEGER',
    'Rich_Media_Event_Id': 'INTEGER',
    'Dbm_Url': 'STRING',
    'Dbm_Cpm_Fee_4_Partner_Currency': 'FLOAT',
    'Dbm_Country_Id': 'STRING',
    'Dbm_Bid_Price_Usd': 'INTEGER',
    'Campaign_End_Date': 'INTEGER',
    'Dbm_Media_Fee_1_Usd': 'FLOAT',
    'Dbm_Operating_System_Id': 'INTEGER',
    'Designated_Market_Area_Dma_Id': 'INTEGER',
    'Placement_Rate': 'INTEGER',
    'Placement': 'STRING',
    'Event_Type': 'STRING',
    'Partner_5_Id': 'STRING',
    'Ad_Type': 'STRING',
    'Segment_Value_1': 'INTEGER',
    'Total_Revenue': 'FLOAT',
    'Dbm_Media_Fee_2_Advertiser_Currency': 'FLOAT',
    'Dbm_Cpm_Fee_3_Advertiser_Currency': 'FLOAT',
    'Browser_Platform': 'STRING',
    'Dbm_Cpm_Fee_2_Usd': 'FLOAT',
    'Dbm_Data_Fees_Usd': 'FLOAT',
    'Google_Click_Id': 'STRING',
    'Dbm_Attributed_Inventory_Source_External_Id': 'STRING',
    'Dbm_Data_Fees_Partner_Currency': 'FLOAT',
    'Custom_Event_Counters': 'INTEGER',
    'Tran_Value': 'STRING',
    'U_Value': 'STRING',
    'Campaign_Start_Date': 'INTEGER',
    'Ad_Type_Id': 'INTEGER',
    'Paid_Search_Keyword_Id': 'INTEGER',
    'Dbm_Media_Cost_Partner_Currency': 'INTEGER',
    'Dbm_Media_Fee_1_Partner_Currency': 'FLOAT',
    'Dbm_Language': 'STRING',
    'Dbm_Adx_Page_Categories': 'STRING',
    'Browser_Platform_Id': 'INTEGER',
    'Floodlight_Variable_Id': 'INTEGER',
    'Dbm_Site_Id': 'INTEGER',
    'Dbm_Cpm_Fee_5_Usd': 'FLOAT',
    'Null_User_Id_Reason_Groups': 'STRING',
    'Dbm_Ad_Position': 'INTEGER',
    'Placement_Cost_Structure': 'STRING',
    'Paid_Search_Legacy_Keyword_Id': 'INTEGER',
    'Dbm_Billable_Cost_Usd': 'INTEGER',
    'Landing_Page_Url_Id': 'INTEGER',
    'Dbm_Isp_Id': 'INTEGER',
    'Paid_Search_Ad_Id': 'INTEGER',
    'Dbm_Browser_Timezone_Offset_Minutes': 'INTEGER',
    'Dbm_Bid_Price_Partner_Currency': 'INTEGER',
    'Rich_Media_Event_Type': 'STRING',
    'Activity_Group_Id': 'INTEGER',
    'Placement_Start_Date': 'INTEGER',
    'Active_View_Viewable_Impressions': 'INTEGER',
    'Dbm_Bid_Price_Advertiser_Currency': 'INTEGER',
    'Dbm_Media_Fee_2_Partner_Currency': 'FLOAT',
    'Dbm_Billable_Cost_Partner_Currency': 'INTEGER',
    'Placement_Total_Booked_Units': 'INTEGER',
    'Dbm_Total_Media_Cost_Usd': 'INTEGER',
    'Creative_Field_Value': 'STRING',
    'Dbm_Cpm_Fee_1_Advertiser_Currency': 'FLOAT',
    'Dbm_Media_Fee_5_Advertiser_Currency': 'FLOAT',
    'Creative_Field_Value_Id': 'INTEGER',
    'Dbm_Cpm_Fee_3_Partner_Currency': 'FLOAT',
    'Dma_Region_Id': 'INTEGER',
    'Partner_4_Id': 'STRING',
    'Activity_Id': 'INTEGER',
    'Paid_Search_Bid_Strategy': 'STRING',
    'City_Id': 'INTEGER',
    'Dbm_Insertion_Order_Id': 'INTEGER',
    'Placement_Id': 'INTEGER',
    'Placement_Strategy_Id': 'INTEGER',
    'Creative_Image_Url': 'STRING',
    'Dbm_Zip_Postal_Code': 'STRING',
    'Site_Directory': 'STRING',
    'Dbm_Auction_Id': 'STRING',
    'Dbm_Device_Type': 'INTEGER',
    'Conversion_Id': 'INTEGER',
    'Creative_Rotation_Type_Id': 'INTEGER',
    'Dbm_Matching_Targeted_Segments': 'STRING',
    'Ad': 'STRING',
    'Advertiser': 'STRING',
    'User_Id': 'STRING',
    'Creative_Field_Id': 'INTEGER',
    'Dbm_Revenue_Partner_Currency': 'FLOAT',
    'Dbm_Attributed_Inventory_Source_Is_Public': 'STRING',
    'Dbm_Net_Speed': 'INTEGER',
    'Site_Directory_Id': 'INTEGER',
    'Creative_End_Date': 'INTEGER',
    'Partner_1_Id': 'STRING',
    'Advertiser_Group': 'STRING',
    'City': 'STRING',
    'Paid_Search_Ad_Group': 'STRING',
    'Dbm_Media_Cost_Advertiser_Currency': 'INTEGER',
    'Dbm_Cpm_Fee_1_Partner_Currency': 'FLOAT',
    'Site': 'STRING',
    'Partner_3_Id': 'STRING',
    'Activity_Sub_Type': 'STRING',
    'Dbm_Line_Item_Id': 'INTEGER',
    'Creative_Field_Name': 'STRING',
    'Dbm_View_State': 'STRING',
    'Dbm_Mobile_Model_Id': 'INTEGER',
    'Creative_Pixel_Size': 'STRING',
    'Operating_System': 'STRING',
    'Dbm_Exchange_Id': 'STRING',
    'Operating_System_Id': 'INTEGER',
    'Dbm_Billable_Cost_Advertiser_Currency': 'INTEGER',
    'Dbm_Data_Fees_Advertiser_Currency': 'FLOAT',
    'Dbm_Cpm_Fee_5_Partner_Currency': 'FLOAT',
    'Ad_Click_Url': 'STRING',
    'Need_Description': 'STRING',
    'Placement_Strategy': 'STRING',
    'Dbm_Cpm_Fee_1_Usd': 'FLOAT',
    'Dbm_Media_Fee_1_Advertiser_Currency': 'FLOAT',
    'Rich_Media_Event_Name': 'STRING',
    'Placement_Comments': 'STRING',
    'Campaign_Id': 'INTEGER',
    'Dbm_Matching_Targeted_Keywords': 'STRING',
    'Ad_Comments': 'STRING',
    'Paid_Search_Bid_Strategy_Id': 'INTEGER',
    'Dbm_State_Region_Id': 'INTEGER',
    'Creative_Last_Modified_Date': 'INTEGER',
    'Creative_Rotation_Type': 'STRING',
    'Activity': 'STRING',
    'Ord_Value': 'STRING',
    'Paid_Search_Match_Type': 'STRING',
    'Dbm_Media_Fee_2_Usd': 'FLOAT',
    'Active_View_Measurable_Impressions': 'INTEGER',
    'Dbm_Creative_Id': 'INTEGER',
    'Impression_Id': 'STRING',
    'Dbm_Media_Fee_4_Advertiser_Currency': 'FLOAT',
    'Activity_Type': 'STRING',
    'Paid_Search_Ad_Group_Id': 'INTEGER',
    'Floodlight_Config_Id': 'INTEGER',
    'Creative_Group2': 'STRING',
    'Paid_Search_Campaign_Id': 'INTEGER',
    'Creative_Group1': 'STRING',
    'State_Region': 'STRING',
    'Placement_End_Date': 'INTEGER',
    'Dbm_Cpm_Fee_2_Partner_Currency': 'FLOAT',
    'Creative_Id': 'INTEGER',
    'Dbm_City_Id': 'INTEGER',
    'Dbm_Designated_Market_Area_Dma_Id': 'INTEGER'
}



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################
"""Handles creation and updates of creative asset association."""

import json

from starthinker.task.traffic.dao import BaseDAO
from starthinker.task.traffic.feed import FieldMap
from starthinker.task.traffic.campaign import CampaignDAO
from starthinker.task.traffic.creative import CreativeDAO


class CreativeAssociationDAO(BaseDAO):
  """Creative Association data access object.

  Inherits from BaseDAO and implements creative association specific logic for
  creating and
  updating creative association.
  """

  def __init__(self, auth, profile_id, is_admin):
    """Initializes CreativeAssociationDAO with profile id and authentication scheme."""
    super(CreativeAssociationDAO, self).__init__(auth, profile_id, is_admin)

    self._id_field = FieldMap.CAMPAIGN_CREATIVE_ASSOCIATION_ID

    self.campaign_dao = CampaignDAO(auth, profile_id, is_admin)
    self.creative_dao = CreativeDAO(auth, profile_id, is_admin)

    self._parent_filter_name = None
    self._parent_filter_field_name = None

  def _api(self, iterate=False):
    """Returns an DCM API instance for this DAO."""
    return super(CreativeAssociationDAO,
                 self)._api(iterate).campaignCreativeAssociations()

  def get(self, feed_item):
    """It is not possible to retrieve creative associations from DCM,

    and they are read-only, so the get method just returns None,
    it does this to avoid errors when this is invoked polimorfically.

    For more information on the get method, refer to BaseDAO.

    Args:
      feed_item: Feed item representing the creative association from the
        Bulkdozer feed.

    Returns:
      None.
    """
    return None

  def process(self, feed_item):
    """Processes a feed item by creating the creative association in DCM.

    Args:
      feed_item: Feed item representing the creative association from the
        Bulkdozer feed.

    Returns:
      The newly created object from DCM.
    """
    if not feed_item.get(FieldMap.CAMPAIGN_CREATIVE_ASSOCIATION_ID, None):
      campaign = self.campaign_dao.get(feed_item, required=True)
      creative = self.creative_dao.get(feed_item, required=True)

      if campaign and creative:
        if campaign:
          feed_item[FieldMap.CAMPAIGN_ID] = campaign['id']
          feed_item[FieldMap.CAMPAIGN_NAME] = campaign['name']

        association = {'creativeId': creative['id']}

        result = self._api().insert(
            profileId=self.profile_id,
            campaignId=campaign['id'],
            body=association).execute()

        feed_item[FieldMap.CAMPAIGN_CREATIVE_ASSOCIATION_ID] = '%s|%s' % (
            campaign['id'], creative['id'])

        return result

    return None
